# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from enum import Enum

security_headers = {
    "Content-Security-Policy": (
        "default-src 'self'; connect-src 'self'; script-src 'unsafe-inline'; "
        "style-src 'unsafe-inline'; img-src 'self' data:; font-src data:;"
    ),
    "X-Frame-Options": "SAMEORIGIN",
    "X-XSS-Protection": "1; mode=block",
    "X-Content-Type-Options": "nosniff",
    "Referrer-Policy": "no-referrer",
}

ADD_MATCH_KEYS = [
    'MaxAbsErr',
    'MinAbsErr',
    'MeanAbsErr',
    'NormAbsErr',
    'MaxRelativeErr',
    'MinRelativeErr',
    'MeanRelativeErr',
    'NormRelativeErr',
]
MAX_FILE_SIZE = 15 * 1024 * 1024 * 1024  # 最大文件大小限制
NPU_PREFIX = 'N___'
BENCH_PREFIX = 'B___'
FILE_NAME_REGEX = r'^[a-zA-Z0-9_\-\.]+$'  # 文件名正则表达式
COLOR_PATTERN = r'^#[0-9A-Fa-f]{6}$'
# 未匹配节点值
UN_MATCHED_VALUE = -1
# 图类型
NPU = 'NPU'
BENCH = 'Bench'
SINGLE = 'Single'

# 前端节点类型
EXPAND_MODULE = 0
UNEXPAND_NODE = 1

# 权限码
PERM_GROUP_WRITE = 0o020
PERM_OTHER_WRITE = 0o002

# 后端节点类型
MODULE = 0
API = 1
MULTI_COLLECTION = 8
API_LIST = 9

# 计算指标
MAX_RELATIVE_ERR = "0"
MIN_RELATIVE_ERR = "1"
MEAN_RELATIVE_ERR = "2"
NORM_RELATIVE_ERR = "3"


class Extension(Enum):
    DB = '.vis.db'
    JSON = '.vis'


class DataType(Enum):
    DB = 'db'
    JSON = 'json'



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves generate dump data function.
"""
import os
import time

import numpy as np

from msprobe.core.common.file_utils import save_numpy_to_bin
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.core.common.log import logger
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException
from msprobe.infer.utils.check.rule import Rule


class DumpData(object):
    """
    Class for generate dump data.
    """

    def __init__(self):
        self.net_output = {}
        pass

    @staticmethod
    def _to_valid_name(name_str):
        return name_str.replace('.', '_').replace('/', '_')

    @staticmethod
    def _check_path_exists(input_path, extentions=None):
        input_path = os.path.realpath(input_path)
        if not os.path.exists(input_path):
            logger.error(f"path '{input_path}' not exists")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

        if extentions and not any([input_path.endswith(extention) for extention in extentions]):
            logger.error(f"path '{input_path}' not ends with extention {extentions}")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

        if not os.access(input_path, os.R_OK):
            logger.error(f"user doesn't have read permission to the file {input_path}.")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

    @staticmethod
    def _check_input_data_path(input_path, inputs_tensor_info):
        if len(inputs_tensor_info) != len(input_path):
            logger.error(
                f"the number of model inputs tensor_info is not equal the number of inputs data, "
                f"inputs tensor_info is: {len(inputs_tensor_info)}, "
                f"inputs data is: {len(input_path)}"
            )
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_DATA_ERROR)

        for cur_path in input_path:
            if not os.path.exists(cur_path):
                logger.error(f"input data path '{cur_path}' not exists")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_PATH_ERROR)

    @staticmethod
    def _generate_random_input_data(save_dir, names, shapes, dtypes):
        inputs_map = {}
        for index, (tensor_name, tensor_shape, tensor_dtype) in enumerate(zip(names, shapes, dtypes)):
            input_data = np.random.random(tensor_shape).astype(tensor_dtype)
            inputs_map[tensor_name] = input_data
            file_name = "input_" + str(index) + ".bin"
            input_data_path = os.path.join(save_dir, file_name)
            save_numpy_to_bin(input_data, input_data_path)
            logger.info(f"save input file name: {file_name}, shape: {input_data.shape}, dtype: {input_data.dtype}")
        return inputs_map

    def generate_dump_data(self):
        pass

    def get_net_output_info(self):
        return self.net_output

    def generate_inputs_data(self):
        pass

    def _generate_dump_data_file_name(self, name_str, node_id):
        return ".".join([self._to_valid_name(name_str), str(node_id), str(round(time.time() * 1e6)), "npy"])

    def _read_input_data(self, input_pathes, names, shapes, dtypes):
        inputs_map = {}
        for input_path, name, shape, dtype in zip(input_pathes, names, shapes, dtypes):
            if dtype == np.float32 and os.path.getsize(input_path) == np.prod(shape) * 2:
                if Rule.input_file().check(input_path, will_raise=True):
                    input_data = np.fromfile(input_path, dtype=np.float16).astype(np.float32)
            else:
                if Rule.input_file().check(input_path, will_raise=True):
                    input_data = np.fromfile(input_path, dtype=dtype)
            if np.prod(input_data.shape) != np.prod(shape):
                cur = input_data.shape
                logger.error(f"input data shape not match, input_path: {input_path}, shape: {cur}, target: {shape}")
                raise AccuracyCompareException(utils.ACCURACY_COMPARISON_INVALID_DATA_ERROR)
            input_data = input_data.reshape(shape)
            inputs_map[name] = input_data
            logger.info(
                f"load input file name: {os.path.basename(input_path)}, "
                f"shape: {input_data.shape}, "
                f"dtype: {input_data.dtype}"
            )
        return inputs_map



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import mindspore
import numpy as np
from mindspore._c_expression import typing
from mindspore.common import dtype as mstype

from msprobe.mindspore.api_accuracy_checker import torch_mindtorch_importer

if torch_mindtorch_importer.is_valid_pt_mt_env:
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import torch
else:
    from msprobe.mindspore.api_accuracy_checker.torch_mindtorch_importer import mindtorch
    import torch

INT8 = "Int8"
UINT8 = "UInt8"
INT16 = "Int16"
UINT16 = "UInt16"
INT32 = "Int32"
UINT32 = "UInt32"
INT64 = "Int64"
UINT64 = "UInt64"
FLOAT16 = "Float16"
FLOAT32 = "Float32"
FLOAT64 = "Float64"
BOOL = "Bool"
BFLOAT16 = "BFloat16"
INT4 = "Int4"

dtype_str_to_ms_dtype = {
    INT8: mstype.int8,
    UINT8: mstype.uint8,
    INT16: mstype.int16,
    UINT16: mstype.uint16,
    INT32: mstype.int32,
    UINT32: mstype.uint32,
    INT64: mstype.int64,
    UINT64: mstype.uint64,
    FLOAT16: mstype.float16,
    FLOAT32: mstype.float32,
    FLOAT64: mstype.float64,
    BOOL: mstype.bool_,
    BFLOAT16: mstype.bfloat16,
    INT4: mstype.qint4x2
}
ms_dtype_to_dtype_str = {value: key for key, value in dtype_str_to_ms_dtype.items()}

dtype_str_to_np_dtype = {
    INT8: np.int8,
    UINT8: np.uint8,
    INT16: np.int16,
    UINT16: np.uint16,
    INT32: np.int32,
    UINT32: np.uint32,
    INT64: np.int64,
    UINT64: np.uint64,
    FLOAT16: np.float16,
    FLOAT32: np.float32,
    FLOAT64: np.float64,
    BOOL: np.bool_
}
np_dtype_to_dtype_str = {value: key for key, value in dtype_str_to_np_dtype.items()}

dtype_str_to_torch_dtype = {
    INT8: torch.int8,
    UINT8: torch.uint8,
    INT16: torch.int16,
    INT32: torch.int32,
    INT64: torch.int64,
    FLOAT16: torch.float16,
    FLOAT32: torch.float32,
    FLOAT64: torch.float64,
    BOOL: torch.bool,
    BFLOAT16: torch.bfloat16,
}
torch_dtype_to_dtype_str = {value: key for key, value in dtype_str_to_torch_dtype.items()}


dtype_str_to_mindtorch_dtype = {
    INT8: mindtorch.int8,
    UINT8: mindtorch.uint8,
    INT16: mindtorch.int16,
    INT32: mindtorch.int32,
    INT64: mindtorch.int64,
    FLOAT16: mindtorch.float16,
    FLOAT32: mindtorch.float32,
    FLOAT64: mindtorch.float64,
    BOOL: mindtorch.bool,
    BFLOAT16: mindtorch.bfloat16,
}
mindtorch_dtype_to_dtype_str = {value: key for key, value in dtype_str_to_mindtorch_dtype.items()}

MINDSPORE_TENSOR_TYPE_STR = "mindspore.Tensor"
BOOL_TYPE_STR = "bool"
INT_TYPE_STR = "int"
FLOAT_TYPE_STR = "float"
SLICE_TYPE_STR = "slice"
TUPLE_TYPE_STR = "tuple"
STR_TYPE_STR = "str"
MINDSPORE_DTYPE_TYPE_STR = "mindspore.dtype"
TORCH_DTYPE_TYPE_STR = "torch.dtype"

api_info_type_str_to_type = {
    MINDSPORE_TENSOR_TYPE_STR: mindspore.Tensor,
    BOOL_TYPE_STR: bool,
    INT_TYPE_STR: int,
    FLOAT_TYPE_STR: float,
    SLICE_TYPE_STR: slice,
    STR_TYPE_STR: str,
    MINDSPORE_DTYPE_TYPE_STR: typing.Type,
}
type_to_api_info_type_str = {value: key for key, value in api_info_type_str_to_type.items()}

DEFAULT_CONSTRUCT_NP_FLOAT_DTYPE = np.float64
DEFAULT_CONSTRUCT_NP_INT_DTYPE = np.float64
DEFAULT_CONSTRUCT_NP_UINT_DTYPE = np.float64

float_dtype_str_list = [
    FLOAT16,
    FLOAT32,
    FLOAT64,
    BFLOAT16,
]

int_dtype_str_list = [
    INT8,
    INT16,
    INT32,
    INT64,
    BOOL,
    INT4,
]

uint_dtype_str_list = [
    UINT8,
    UINT16,
    UINT32,
    UINT64,
]



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
import multiprocessing
from dataclasses import dataclass

import numpy as np
import pandas as pd
from tqdm import tqdm

from msprobe.core.common.file_utils import check_file_or_directory_path, create_directory, load_npy, save_excel
from msprobe.core.common.log import logger
from msprobe.core.common.utils import check_process_num


@dataclass
class CompareResult:
    max_abs_error: float
    max_relative_error: float
    same_percentage: float
    first_mismatch_index: int
    percentage_within_thousandth: float
    percentage_within_hundredth: float


class SingleComparator:
    result_header = [
        'step', 
        'rank', 
        'micro_step', 
        'id', 
        'shape1', 
        'shape2', 
        '相同元素百分比(%)', 
        '首个不匹配元素索引', 
        '最大绝对误差', 
        '最大相对误差', 
        '误差在千分之一内元素占比(%)', 
        '误差在百分之一内元素占比(%)'
    ]

    @classmethod
    def compare(cls, dir1, dir2, output_path="./msprobe_compare_output", num_processes=8):
        data_dir1 = os.path.join(dir1, "data")
        data_dir2 = os.path.join(dir2, "data")
        check_file_or_directory_path(data_dir1, isdir=True)
        check_file_or_directory_path(data_dir2, isdir=True)
        check_process_num(num_processes)
        # 确保输出目录存在，如果不存在则创建
        if not os.path.exists(output_path):
            create_directory(output_path)
        cls.compare_data(data_dir1, data_dir2, output_path, num_processes)

    @classmethod
    def compare_arrays(cls, array1, array2) -> CompareResult:
        """
        比较两个NumPy数组，计算最大绝对误差、最大相对误差和相同元素的百分比
        """
        # 计算每个维度上的最小尺寸
        if array1.ndim != array2.ndim:
            array1 = array1.flatten()
            array2 = array2.flatten()
        min_shape = [min(s1, s2) for s1, s2 in zip(array1.shape, array2.shape)]
        # 截取数组到相同的形状
        sliced_array1 = array1[tuple(slice(0, s) for s in min_shape)]
        sliced_array2 = array2[tuple(slice(0, s) for s in min_shape)]

        abs_error = np.abs(sliced_array1 - sliced_array2)
        max_abs_error = np.max(abs_error)
        
        # 计算相对误差，处理分母为零的情况
        with np.errstate(divide='ignore', invalid='ignore'):
            relative_error = np.abs(sliced_array1 - sliced_array2) / \
                np.maximum(np.abs(sliced_array1), np.abs(sliced_array2))
        relative_error = np.nan_to_num(relative_error)
        max_relative_error = np.max(relative_error)

        same_elements = np.sum(sliced_array1 == sliced_array2)
        total_elements = sliced_array1.size
        same_percentage = (same_elements / total_elements) * 100

        # 展平数组
        flat_array1 = sliced_array1.flatten()
        flat_array2 = sliced_array2.flatten()

        # 计算从第几个元素开始对不上
        mismatch_indices = np.nonzero(flat_array1 != flat_array2)[0]
        first_mismatch_index = mismatch_indices[0] if mismatch_indices.size > 0 else None

        # 计算误差在千分之一内的元素占比
        threshold = 0.001 * np.maximum(np.abs(sliced_array1), np.abs(sliced_array2))
        error_within_thousandth = np.sum(abs_error <= threshold)
        percentage_within_thousandth = (error_within_thousandth / total_elements) * 100

        # 计算误差在百分之一内的元素占比
        threshold = 0.01 * np.maximum(np.abs(sliced_array1), np.abs(sliced_array2))
        error_within_hundredth = np.sum(abs_error <= threshold)
        percentage_within_hundredth = (error_within_hundredth / total_elements) * 100

        return CompareResult(
            max_abs_error,
            max_relative_error,
            same_percentage,
            first_mismatch_index,
            percentage_within_thousandth,
            percentage_within_hundredth
        )

    @classmethod
    def get_steps(cls, tag_path):
        for step_folder in os.listdir(tag_path):
            if step_folder.startswith('step'):
                try:
                    step = int(step_folder[4:])
                except Exception as e:
                    raise RuntimeError(f"parse step number error") from e
                yield step, os.path.join(tag_path, step_folder)

    @classmethod
    def get_ranks(cls, step_path):
        for rank_folder in os.listdir(step_path):
            if rank_folder.startswith('rank'):
                try:
                    rank = int(rank_folder[4:])
                except Exception as e:
                    raise RuntimeError(f"parse rank number error") from e
                yield rank, os.path.join(step_path, rank_folder)

    @classmethod
    def get_micro_steps(cls, rank_path):
        for micro_step_folder in os.listdir(rank_path):
            if micro_step_folder.startswith('micro_step'):
                try:
                    micro_step = int(micro_step_folder[10:])
                except Exception as e:
                    raise RuntimeError(f"parse nicro_step number error") from e
                yield micro_step, os.path.join(rank_path, micro_step_folder)
            else:
                yield 0, rank_path

    @classmethod
    def get_arrays(cls, micro_step_path):
        for file in os.listdir(micro_step_path):
            if file.endswith('.npy'):
                try:
                    parts = file.rsplit('.', 2)
                    if len(parts) > 1 and parts[-2].isdigit():
                        array_id = int(parts[-2])
                    else:
                        array_id = 0
                except ValueError:
                    array_id = 0
                yield array_id, os.path.join(micro_step_path, file)

    @classmethod
    def get_array_paths(cls, dir_path):
        """
        获取目录中所有符合结构的NumPy数组文件路径
        """
        array_paths = {}
        if not os.path.exists(dir_path):
            return array_paths
        for tag in os.listdir(dir_path):
            tag_path = os.path.join(dir_path, tag)
            if not os.path.isdir(tag_path):
                continue
            for step, step_path in cls.get_steps(tag_path):
                for rank, rank_path in cls.get_ranks(step_path):
                    for item in os.listdir(rank_path):
                        next_path = os.path.join(rank_path, item)
                        if re.match(r"micro_step(\d+)", item):
                            micro_step = re.match(r"micro_step(\d+)", item).group(1)
                            for array_id, array_path in cls.get_arrays(next_path):
                                array_paths.setdefault(tag, []).append(
                                    (step, rank, int(micro_step), array_id, array_path))
                        elif re.match(r"\w{1,100}_(\d{1,100})\.npy", item):
                            array_id = re.match(r"\w{1,100}_(\d{1,100})\.npy", item).group(1)
                            array_paths.setdefault(tag, []).append((step, rank, 0, int(array_id), next_path))
                        else:
                            array_paths.setdefault(tag, []).append((step, rank, 0, 0, next_path))
        return array_paths

    @classmethod
    def compare_single_tag(cls, tag, array_paths1, array_paths2, output_dir):
        data = []
        paths1 = array_paths1.get(tag, [])
        paths2 = array_paths2.get(tag, [])
        path_dict1 = {(step, rank, micro_step, array_id): path for step, rank, micro_step, array_id, path in paths1}
        path_dict2 = {(step, rank, micro_step, array_id): path for step, rank, micro_step, array_id, path in paths2}
        common_keys = set(path_dict1.keys()) & set(path_dict2.keys())
        for key in common_keys:
            try:
                array1 = load_npy(path_dict1[key])
                array2 = load_npy(path_dict2[key])
                result = cls.compare_arrays(array1, array2)
                step, rank, micro_step, array_id = key
                data.append([
                    step, rank, micro_step, array_id,
                    list(array1.shape), list(array2.shape),
                    result.same_percentage,
                    result.first_mismatch_index,
                    result.max_abs_error,
                    result.max_relative_error,
                    result.percentage_within_thousandth,
                    result.percentage_within_hundredth
                ])
            except Exception as e:
                logger.error(f"Error comparing {path_dict1[key]} and {path_dict2[key]}: {e}")

        try:
            df = pd.DataFrame(data, columns=SingleComparator.result_header)
            df = df.sort_values(by=['step', 'rank', 'micro_step', 'id'])
            # 构建输出文件的完整路径
            output_file_path = os.path.join(output_dir, f'{tag}.xlsx')
            save_excel(output_file_path, df)
        except Exception as e:
            logger.error(f"Error processing tag {tag}: {e}")

    @classmethod
    def compare_data(cls, dir1, dir2, output_dir, num_processes=8):
        """
        比较两个目录中的NumPy数组文件，并将结果保存到指定目录的Excel文件中
        """

        array_paths1 = cls.get_array_paths(dir1)
        array_paths2 = cls.get_array_paths(dir2)

        all_tags = set(array_paths1.keys()) | set(array_paths2.keys())

        with multiprocessing.Pool(processes=num_processes) as pool:
            args = [(tag, array_paths1, array_paths2, output_dir) for tag in all_tags]
            try:
                results = pool.starmap_async(cls.compare_single_tag, args)
                with tqdm(total=len(all_tags), desc="Processing data") as pbar:
                    while not results.ready():
                        pbar.n = len(all_tags) - results._number_left
                        pbar.refresh()
                    results.wait()
                    results.get()
            except Exception as e:
                logger.error(f"Multiprocessing error: {e}")
            finally:
                pool.close()
                pool.join()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import namedtuple
import re
import numpy as np
import torch
try:
    import torch_npu
except ImportError:
    current_device = "cuda"
    from torch.cuda.amp import autocast
    IS_GPU = True
else:
    current_device = "npu"
    from torch_npu.npu.amp import autocast
    IS_GPU = False

from msprobe.core.common.const import FileCheckConst, Const, CompareConst
from msprobe.core.common.file_utils import FileChecker
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException
from msprobe.pytorch.dump.api_dump.api_register import ApiTemplate, get_api_register
from msprobe.pytorch.api_accuracy_checker.common.utils import is_dtype_fp8, is_hifloat8_tensor


hf_32_standard_api = ["conv1d", "conv2d"]
not_detach_set = {'resize_', 'resize_as_', 'set_', 'transpose_', 't_', 'squeeze_', 'unsqueeze_'}
not_raise_dtype_set = {'type_as'}

PRECISION_MAPPING = {
    torch.float16: torch.float32,
    torch.bfloat16: torch.float32,
    torch.float32: torch.float64
}


CpuParams = namedtuple("CpuArgs", ["cpu_args", "cpu_kwargs", "autocast_dtype", "is_autocast"])
ExecParams = namedtuple("ExecParams", ["api_type", "api_name", "device", "args", "kwargs", 
                                       "is_autocast", "autocast_dtype"])


class BackwardMessage:
    MULTIPLE_BACKWARD_MESSAGE = "Multiple backward is not supported."
    UNSUPPORT_BACKWARD_MESSAGE = "function with out=... arguments don't support automatic differentiation, " \
                                  "skip backward."
    NO_BACKWARD_RESULT_MESSAGE = "This API does not have backward input data, skip backward."
    UNSUPPORT_API_MESSAGE = "This API does not support backward ut, skip backward."


class UtDataInfo:
    def __init__(self, bench_grad, device_grad, device_output, bench_output, grad_in, in_fwd_data_list,
                 backward_message, rank=0, is_fp8=False):
        self.bench_grad = bench_grad
        self.device_grad = device_grad
        self.device_output = device_output
        self.bench_output = bench_output
        self.grad_in = grad_in
        self.in_fwd_data_list = in_fwd_data_list
        self.backward_message = backward_message
        self.rank = rank
        self.is_fp8 = is_fp8


def get_validated_result_csv_path(result_csv_path, mode):
    if mode not in ['result', 'detail']:
        raise ValueError("The csv mode must be result or detail")
    result_csv_path_checker = FileChecker(result_csv_path, FileCheckConst.FILE, ability=FileCheckConst.READ_WRITE_ABLE,
                                          file_type=FileCheckConst.CSV_SUFFIX)
    validated_result_csv_path = result_csv_path_checker.common_check()
    if mode == 'result':
        result_csv_name = os.path.basename(validated_result_csv_path)
        pattern = r"^accuracy_checking_result_\d{14}\.csv$"
        if not re.match(pattern, result_csv_name):
            raise ValueError("When continue acc_check, please do not modify the result csv name.")
    return validated_result_csv_path


def get_validated_details_csv_path(validated_result_csv_path):
    result_csv_name = os.path.basename(validated_result_csv_path)
    details_csv_name = result_csv_name.replace('result', 'details')
    details_csv_path = os.path.join(os.path.dirname(validated_result_csv_path), details_csv_name)
    details_csv_path_checker = FileChecker(details_csv_path, FileCheckConst.FILE,
                                           ability=FileCheckConst.READ_WRITE_ABLE, file_type=FileCheckConst.CSV_SUFFIX)
    validated_details_csv_path = details_csv_path_checker.common_check()
    return validated_details_csv_path


def exec_api(exec_params):
    api_type = exec_params.api_type
    api_name = exec_params.api_name
    device = exec_params.device
    args = exec_params.args
    kwargs = exec_params.kwargs
    is_autocast = exec_params.is_autocast
    autocast_dtype = exec_params.autocast_dtype
    out = None

    prefix_map = Const.API_DATA_PREFIX.get(Const.PT_FRAMEWORK, {})
    if not prefix_map or api_type not in prefix_map.values() or \
        api_type not in (
            Const.FUNCTIONAL_API_TYPE_PREFIX,
            Const.TENSOR_API_TYPE_PREFIX,
            Const.TORCH_API_TYPE_PREFIX,
            Const.ATEN_API_TYPE_PREFIX,
            Const.NPU_API_TYPE_PREFIX
    ):
        return out

    api_register = get_api_register()
    api_register.initialize_hook(None)
    api_func_type = list(prefix_map.keys())[list(prefix_map.values()).index(api_type)]
    api_func = api_register.ori_api_attr.get(Const.PT_FRAMEWORK + Const.SEP + api_func_type, {}).get(api_name)

    torch_api = ApiTemplate(api_name, api_func, api_type, None, need_hook=False, device=device)
    if is_autocast:
        with autocast(dtype=autocast_dtype):
            out = torch_api.forward(*args, **kwargs)
    else:
        out = torch_api.forward(*args, **kwargs)
    return out


def deal_detach(arg, to_detach=True):
    return arg.detach() if to_detach else arg


def raise_bench_data_dtype(api_name, arg, raise_dtype=None):
    '''
    将标杆数据的dtype转换为raise_dtype
    输入：
        api_name：api名称
        arg：标杆输入
        raise_dtype：需要转换的dtype
    输出： 
        arg: 转换dtype的标杆输入
    '''
    if is_hifloat8_tensor(arg):
        return hif8_to_fp32(arg)
    if is_dtype_fp8(arg.dtype):
        return fp8_to_fp32(arg)
    if api_name in hf_32_standard_api and arg.dtype == torch.float32:
        return arg
    if raise_dtype is None or arg.dtype not in PRECISION_MAPPING or raise_dtype == arg.dtype:
        return arg
    return arg.type(raise_dtype)


def generate_device_params(input_args, input_kwargs, need_backward, api_name):
    is_fp8 = False
    
    def recursive_arg_to_device(arg_in, to_detach, depth=0):
        nonlocal is_fp8
        if depth > Const.MAX_DEPTH:
            logger.error("The depth of arg_in is too large, please check the arg_in.")
            raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
        if isinstance(arg_in, (list, tuple)):
            return type(arg_in)(recursive_arg_to_device(arg, to_detach, depth=depth + 1) for arg in arg_in)
        elif isinstance(arg_in, torch.Tensor):
            if is_dtype_fp8(arg_in.dtype) or is_hifloat8_tensor(arg_in):
                is_fp8 = True
            if need_backward and arg_in.requires_grad:
                arg_in = deal_detach(arg_in.clone(), to_detach).to(current_device).requires_grad_()
                temp_arg_in = arg_in * 1
                arg_in = temp_arg_in.type_as(arg_in)
                arg_in.retain_grad()
                return arg_in
            else:
                return deal_detach(arg_in.clone(), to_detach).to(current_device)
        else:
            return arg_in

    is_detach = api_name not in not_detach_set
    device_args = recursive_arg_to_device(input_args, is_detach)
    device_kwargs = \
        {key: recursive_arg_to_device(value, key != "out" and is_detach) for key, value in input_kwargs.items()}
    return device_args, device_kwargs, is_fp8


def generate_cpu_params(input_args, input_kwargs, need_backward, api_name):
    def recursive_arg_to_cpu(arg_in, to_detach, raise_dtype=None, depth=0):
        if depth > Const.MAX_DEPTH:
            logger.error("The depth of arg_in is too large, please check the arg_in.")
            raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
        if isinstance(arg_in, (list, tuple)):
            return type(arg_in)(recursive_arg_to_cpu(arg, to_detach, raise_dtype=raise_dtype, depth=depth + 1)
                                for arg in arg_in)
        elif isinstance(arg_in, torch.Tensor):
            if need_backward and arg_in.requires_grad:
                arg_in = deal_detach(raise_bench_data_dtype(
                                     api_name, arg_in.clone(), raise_dtype=raise_dtype), to_detach).requires_grad_()
                temp_arg_in = arg_in * 1
                arg_in = temp_arg_in.type_as(arg_in)
                arg_in.retain_grad()
                return arg_in
            else:
                return deal_detach(raise_bench_data_dtype(api_name, arg_in.clone(), raise_dtype=raise_dtype), to_detach)
        else:
            return arg_in

    def is_tensor_with_raise_precision(arg_in, check_kwargs=False):
        if arg_in.dtype in PRECISION_MAPPING:
            return True
        if check_kwargs and arg_in.dtype in [torch.half, torch.bfloat16]:
            return True
        return False

    def recursive_find_dtypes(arg_in, kwargs=None, check_kwargs=False, depth=0):
        if depth > Const.MAX_DEPTH:
            logger.error("The depth of arg_in is too large, please check the arg_in.")
            raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
        if isinstance(arg_in, (list, tuple)):
            return set().union(*tuple(recursive_find_dtypes(arg, kwargs, check_kwargs=check_kwargs, depth=depth + 1) for
                                      arg in arg_in))
        elif isinstance(arg_in, torch.Tensor) and is_tensor_with_raise_precision(arg_in, check_kwargs):
            return set([arg_in.dtype])
        elif isinstance(arg_in, dict) and check_kwargs:
            return set().union(*tuple(recursive_find_dtypes(v, kwargs, check_kwargs=True, depth=depth + 1) for
                                      v in arg_in.values()))
        return set()

    raise_dtype = None
    autocast_dtype = None
    is_autocast = False
    need_raise_dtypes = recursive_find_dtypes(input_args)
    need_raise_dtypes.update(recursive_find_dtypes(input_kwargs, check_kwargs=True))
    if len(need_raise_dtypes) == 1:
        origin_dtype = need_raise_dtypes.pop()
        raise_dtype = PRECISION_MAPPING.get(origin_dtype, torch.float32)
        autocast_dtype = origin_dtype
        
    elif len(need_raise_dtypes) >= 2:
        raise_dtype = torch.float32
        need_raise_dtypes.discard(torch.float32)
        autocast_dtype = need_raise_dtypes.pop()
        is_autocast = True

    raise_dtype = None if api_name in not_raise_dtype_set else raise_dtype
    is_detach = api_name not in not_detach_set
    cpu_args = recursive_arg_to_cpu(input_args, is_detach, raise_dtype=raise_dtype)
    cpu_kwargs = {key: recursive_arg_to_cpu(value, key != "out" and is_detach, raise_dtype=raise_dtype) for
                  key, value in input_kwargs.items()}
    cpu_params = CpuParams(cpu_args, cpu_kwargs, autocast_dtype, is_autocast)
    return cpu_params


def record_skip_info(api_full_name, compare, compare_alg_results):
    result_info = (api_full_name, CompareConst.SKIP, CompareConst.SKIP, [compare_alg_results], None, 0)
    compare.record_results(result_info)


def is_unsupported_api(api_name, is_overflow_check=False):
    split_name = api_name.split(Const.SEP)[0]
    unsupport_type_list = [Const.DISTRIBUTED, Const.MINDSPEED_API_TYPE_PREFIX]
    flag = (split_name in unsupport_type_list) or (is_overflow_check and split_name == Const.NPU)
    if flag:
        logger.info(f"{split_name} api is not supported for acc_check. SKIP.")
    return flag


def fp8_to_fp32(x):
    """
    将FP8格式的张量转换为FP32格式，保持原始FP8的表示范围
    使用纯PyTorch操作替代NumPy位运算
    
    参数:
    x: 输入的FP8张量，可以是torch.float8_e4m3fn或torch.float8_e5m2类型
    
    返回:
    torch.Tensor: 转换后的FP32张量，保持原始FP8的表示范围
    """
    if x.dtype == torch.float8_e4m3fn:
        # E4M3FN格式：1符号+4指数+3尾数，偏置7
        # 位布局：SEEEEEMM
        
        # 将FP8值视为无符号整数进行位操作
        x_int = x.view(torch.uint8)
        
        # 提取符号位、指数位和尾数位
        sign_bits = (x_int & 0x80) >> 7  # 最高位是符号位
        exp_bits = (x_int & 0x78) >> 3   # 接下来4位是指数位
        mantissa_bits = x_int & 0x07     # 最后3位是尾数位
        
        # 处理规格化数和非规格化数
        is_normal = exp_bits != 0
        
        # 计算FP32的指数部分（偏置127）
        fp32_exp = torch.where(
            is_normal,
            (exp_bits - 7 + 127).to(torch.int32),  # 规格化数：指数 = 原始指数 + 120
            torch.tensor(0, dtype=torch.int32, device=x.device)  # 非规格化数：指数为0
        )
        
        # 计算FP32的尾数部分
        # 规格化数：隐含1，尾数 = 1.0 + 原始尾数 * 2^(-3)
        # 非规格化数：无隐含1，尾数 = 0.0 + 原始尾数 * 2^(-3)
        fp32_mantissa = torch.where(
            is_normal,
            1.0 + mantissa_bits.to(torch.float32) / 8.0,  # 2^(-3) = 1/8
            mantissa_bits.to(torch.float32) / 8.0
        )
        
        # 计算符号值 (-1)^sign
        sign_value = torch.pow(-1.0, sign_bits.to(torch.float32))
        
        # 计算最终FP32值
        # 规格化数：value = (-1)^sign * (1.0 + mantissa/8) * 2^(exp - 7)
        # 非规格化数：value = (-1)^sign * (mantissa/8) * 2^(-6)
        fp32_result = sign_value * fp32_mantissa * torch.pow(2.0, fp32_exp - 127)
        
        return fp32_result
    
    elif x.dtype == torch.float8_e5m2:
        # E5M2格式：1符号+5指数+2尾数，偏置15
        # 位布局：SEEEEEEM
        
        # 将FP8值视为无符号整数进行位操作
        x_int = x.view(torch.uint8)
        
        # 提取符号位、指数位和尾数位
        sign_bits = (x_int & 0x80) >> 7  # 最高位是符号位
        exp_bits = (x_int & 0x7C) >> 2   # 接下来5位是指数位
        mantissa_bits = x_int & 0x03     # 最后2位是尾数位
        
        # 处理规格化数和非规格化数
        is_normal = exp_bits != 0
        
        # 计算FP32的指数部分（偏置127）
        fp32_exp = torch.where(
            is_normal,
            (exp_bits - 15 + 127).to(torch.int32),  # 规格化数：指数 = 原始指数 + 112
            torch.tensor(0, dtype=torch.int32, device=x.device)  # 非规格化数：指数为0
        )
        
        # 计算FP32的尾数部分
        # 规格化数：隐含1，尾数 = 1.0 + 原始尾数 * 2^(-2)
        # 非规格化数：无隐含1，尾数 = 0.0 + 原始尾数 * 2^(-2)
        fp32_mantissa = torch.where(
            is_normal,
            1.0 + mantissa_bits.to(torch.float32) / 4.0,  # 2^(-2) = 1/4
            mantissa_bits.to(torch.float32) / 4.0
        )
        
        # 计算符号值 (-1)^sign
        sign_value = torch.pow(-1.0, sign_bits.to(torch.float32))
        
        # 计算最终FP32值
        fp32_result = sign_value * fp32_mantissa * torch.pow(2.0, fp32_exp - 127)
        
        return fp32_result
    
    else:
        raise ValueError(f"Unsupported dtype: {x.dtype}. Expected torch.float8_e4m3fn or torch.float8_e5m2.")


def hif8_to_fp32(x):
    """
    将HiFloat8格式的张量转换为FP32格式，保持原始HiFloat8的表示范围
    使用纯PyTorch操作替代NumPy位运算    
    
    参数:
    x: 输入的HiFloat8张量，可以是torch_npu.HiFloat8Tensor类型
    
    返回:
    torch.Tensor: 转换后的FP32张量，保持原始HiFloat8的表示范围
    """
    requires_grad = x.requires_grad
    x = x.cpu().detach().numpy()
    x = np.array(x)  # 确保输入是numpy数组
    
    # 创建结果数组，保持与输入相同的形状
    res = np.zeros_like(x, dtype=np.float32)
    
    # 获取输入张量的所有维度
    dimensions = x.shape
    # 计算总元素数量
    total_elements = np.prod(dimensions)
    
    # 遍历每个元素
    for idx in range(total_elements):
        # 将一维索引转换为多维索引
        multi_indices = np.unravel_index(idx, dimensions)
        z = x[multi_indices]
        
        # 处理特殊值
        if np.isnan(z) or np.isinf(z):
            res[multi_indices] = z
            continue
            
        # 提取符号位
        s = 1.0 if z >= 0 else -1.0
        tmp = abs(z)
        
        # 处理零值
        if tmp == 0:
            res[multi_indices] = 0.0
            continue
            
        # 确定指数范围和尾数位数
        exponent = np.floor(np.log2(tmp + 1e-100))  # 添加小常量避免log2(0)
        eabs = abs(exponent)
        
        # 根据指数范围确定尾数位数和还原规则
        if eabs <= 3:       # 3-bit Mantissa
            mantissa = (tmp / (2.0 ** exponent)) * 8.0  # 还原尾数部分
            res[multi_indices] = s * (mantissa / 8.0) * (2.0 ** exponent)
        elif eabs <= 7:     # 2-bit Mantissa
            mantissa = (tmp / (2.0 ** exponent)) * 4.0
            res[multi_indices] = s * (mantissa / 4.0) * (2.0 ** exponent)
        elif eabs <= 15:    # 1-bit Mantissa
            mantissa = (tmp / (2.0 ** exponent)) * 2.0
            res[multi_indices] = s * (mantissa / 2.0) * (2.0 ** exponent)
        else:               # 0-bit Mantissa
            res[multi_indices] = s * (2.0 ** exponent)
    
    res = torch.from_numpy(res)
    if requires_grad:
        res = res.requires_grad_()
    return res



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
import math
import zlib
from dataclasses import dataclass
import multiprocessing

import numpy as np
import pandas as pd

from msprobe.core.common.const import Const, CompareConst, FileCheckConst
from msprobe.core.common.utils import CompareException, check_regex_prefix_format_valid, logger, \
    safe_get_value, is_module_available
from msprobe.core.common.file_utils import check_file_or_directory_path, load_json

json_file_mapping = {
    Const.DUMP_JSON_FILE: "dump.json",
    Const.DEBUG_JSON_FILE: "debug.json",
    Const.STACK_JSON_FILE: "stack.json"
}


def extract_json(dirname, json_file_type):
    json_path = ''
    for filename in os.listdir(dirname):
        target_file_name = json_file_mapping.get(json_file_type)
        if target_file_name is None:
            logger.error(f'extract_json failed, invalid json_file_type: {json_file_type}.')
            raise CompareException(CompareException.INVALID_KEY_ERROR)
        if filename == target_file_name:
            json_path = os.path.join(dirname, filename)
            break

    # Provide robustness on invalid directory inputs
    if not json_path:
        if json_file_type == Const.STACK_JSON_FILE:
            logger.warning(f'stack.json is not found in dump dir {dirname}.')
        elif json_file_type == Const.DUMP_JSON_FILE:
            logger.error(f'dump.json is not found in dump dir {dirname}.')
        elif json_file_type == Const.DEBUG_JSON_FILE:
            logger.warning(f'debug.json is not found in dump dir {dirname}.')

    return json_path


def set_stack_json_path(input_param):
    npu_data_dir = os.path.dirname(input_param.get("npu_path"))
    stack_path = extract_json(npu_data_dir, json_file_type=Const.STACK_JSON_FILE)
    input_param["stack_path"] = stack_path if stack_path else None
    return bool(stack_path), input_param


def check_and_return_dir_contents(dump_dir, prefix):
    """
    check the given dump dir and validate files in dump dir by using the given prefix patterns to build a
    pattern: ^{prefix}(?:0|[1-9][0-9]*)?$

    Args:
        dump_dir (str): dump dir
        prefix (str): prefix for the patterns, prefix should be less than 20 characters and alphanumeric/-/_ only

    Returns:
        content [list]: dir contents
    Raises:
        CompareException: invalid path
        ValueError: prefix not match the patterns

    """
    check_regex_prefix_format_valid(prefix)
    check_file_or_directory_path(dump_dir, True)
    contents = os.listdir(dump_dir)
    pattern = re.compile(rf'^{prefix}(?:0|[1-9][0-9]*)?$')
    for name in contents:
        if not pattern.match(name):
            logger.error(
                f"dump_dir contains '{name}'. Expected '{prefix}'. This name is not in the format of dump "
                f"output. Please check and delete irrelevant files in {dump_dir} and try again."
            )
            raise CompareException(CompareException.INVALID_PATH_ERROR)
    return contents


def read_op(op_data, op_name):
    if not isinstance(op_name, str):
        logger.error(f"api name error: {op_name} is not a string, please check.")
        raise CompareException(CompareException.INVALID_API_NAME_ERROR)
    split_name = op_name.split(Const.SEP)
    if split_name[-1] == Const.DEBUG:
        op_parsed_list = op_item_parse(op_data, op_name, Const.DEBUG)
    # 倒序校验parameters_grad，先倒数1，再倒数2，排除因为parameters_grad在模块名中而误判的风险。-1适配老版，-2适配新版。
    elif split_name[-1] == Const.PARAMS_GRAD or (len(split_name) >= 2 and split_name[-2] == Const.PARAMS_GRAD):
        op_parsed_list = op_item_parse(op_data, op_name, Const.PARAMS_GRAD)
    else:
        op_parsed_list = []
        for name in CompareConst.IO_NAME_MAPPING:
            if name in op_data:
                op_parsed_list.extend(op_item_parse(op_data[name], op_name + CompareConst.IO_NAME_MAPPING[name], name))
    return op_parsed_list


def op_item_parse(op_data, op_name: str, state: str, depth: int = 0) -> list:
    if state == Const.INPUT_ARGS or state == Const.INPUT_KWARGS:
        state = Const.INPUT
    default_item = {
        'full_op_name': op_name,
        Const.TYPE: None,
        Const.MAX: None,
        Const.MIN: None,
        Const.MEAN: None,
        Const.NORM: None,
        Const.DTYPE: None,
        Const.SHAPE: None,
        Const.MD5: None,
        Const.VALUE: None,
        Const.DATA_NAME: '-1',
        Const.STATE: state,
        Const.REQ_GRAD: None
    }

    if depth > Const.MAX_DEPTH:
        logger.error(f'parse of api/module of {op_name} exceeds the recursion limit.')
        raise CompareException(CompareException.RECURSION_LIMIT_ERROR)

    if op_data is None:
        return [default_item]
    elif not op_data:
        return []

    item_list = []
    if isinstance(op_data, list):
        for i, data in enumerate(op_data):
            if Const.PARAMS_GRAD not in op_name.split(Const.SEP):
                item_list.extend(op_item_parse(data, op_name + Const.SEP + str(i), state, depth + 1))
            else:
                item_list.extend(op_item_parse(data, op_name, state, depth + 1))
    elif isinstance(op_data, dict):
        if is_p2pop_leaf_data(op_data):
            p2pop_item = {}
            for key in ['class_type', 'op', 'peer', 'tag', 'group_id']:
                p2pop_item[key] = op_data.get(key)
            op_data = op_data.get('tensor')
            if isinstance(op_data, dict):
                op_item = gen_op_item(op_data, op_name, state)
            else:
                op_item = default_item
            op_item.update(p2pop_item)
            return [op_item]
        if is_leaf_data(op_data):
            return [gen_op_item(op_data, op_name, state)]
        for sub_name, sub_data in op_data.items():
            item_list.extend(op_item_parse(sub_data, op_name + Const.SEP + str(sub_name), state, depth + 1))
    return item_list


def is_p2pop_leaf_data(op_data):
    return op_data.get('class_type') == 'torch.distributed.P2POp'


def is_leaf_data(op_data):
    return 'type' in op_data and isinstance(op_data['type'], str)


def gen_op_item(op_data, op_name, state):
    op_item = {}
    op_item.update({key: str(value) if isinstance(value, bool) else value for key, value in op_data.items()})
    data_name = op_data.get(Const.DATA_NAME) if op_data.get(Const.DATA_NAME) else '-1'  # 如果是""也返回-1
    op_item[Const.DATA_NAME] = data_name
    op_item['full_op_name'] = data_name.rsplit(Const.SEP, 1)[0] if data_name != '-1' else op_name
    op_item[Const.STATE] = state
    if Const.REQ_GRAD not in op_item:
        op_item[Const.REQ_GRAD] = None

    # 补齐统计量字段
    params = [Const.MAX, Const.MIN, Const.MEAN, Const.NORM]
    for i in params:
        if i not in op_item:
            op_item[i] = None

    # special cases
    if not op_item.get('dtype'):
        if op_item.get('type') == 'torch.Size':
            op_item['dtype'] = op_data.get('type')
            op_item['shape'] = str(op_data.get('value'))
        elif op_item.get('type') == 'slice':
            op_item['dtype'] = op_data.get('type')
            op_item['shape'] = str(np.shape(np.array(op_data.get('value'))))
        elif op_item.get('type') == 'ellipsis':
            op_item['dtype'] = op_data.get('type')
            op_item['shape'] = '[]'
            for i in params:
                op_item[i] = op_data.get('value')
        elif op_name.split(Const.SEP)[-1] in ['src', 'dst', 'group_src', 'group_dst']:
            op_item['dtype'] = op_data.get('type')
            op_item['shape'] = '[]'
            for i in params:
                op_item[i] = str(op_data.get('value'))
            op_item['md5'] = str(op_data.get('value'))
        elif op_item.get('type') == 'torch.ProcessGroup':
            op_item['dtype'] = op_data.get('type')
            op_item['shape'] = '[]'
            for i in params:
                op_item[i] = str(op_data.get('group_ranks'))
            op_item['md5'] = str(op_data.get('group_ranks'))
        else:
            op_item['dtype'] = str(type(op_data.get('value')))
            op_item['shape'] = '[]'
            for i in params:
                op_item[i] = op_data.get('value')
    if not op_item.get('md5'):
        op_item['md5'] = f"{zlib.crc32(str(op_data.get('value', '')).encode()):08x}"

    return op_item


@dataclass
class ApiItemInfo:
    name: str
    struct: tuple
    stack_info: list


def merge_tensor(tensor_list, dump_mode):
    keys = [
        CompareConst.OP_NAME,
        CompareConst.INPUT_STRUCT,
        CompareConst.KWARGS_STRUCT,
        CompareConst.OUTPUT_STRUCT,
        CompareConst.PARAMS_STRUCT,
        CompareConst.PARAMS_GRAD_STRUCT,
        CompareConst.DEBUG_STRUCT,
        Const.SUMMARY,
        Const.STACK_INFO,
        Const.STATE,
        Const.REQ_GRAD
    ]
    op_dict = {key: [] for key in keys}

    if dump_mode == Const.ALL:
        op_dict[Const.DATA_NAME] = []

    for tensor in tensor_list:
        # A dict(len=2) with 'full_op_name' and 'full_info' is added to the tensor only if self.stack_mode is True
        if len(tensor) == 2:
            op_dict[Const.STACK_INFO].append(tensor.get('full_info'))
            break

        op_dict[CompareConst.OP_NAME].append(tensor.get('full_op_name'))
        state = tensor.get(Const.STATE)
        op_dict[Const.STATE].append(state)
        op_dict[Const.REQ_GRAD].append(tensor.get(Const.REQ_GRAD))

        struct_key = CompareConst.STATE_TO_STRUCT_MAPPING.get(state)
        if not struct_key:
            continue
        if dump_mode == Const.MD5:
            op_dict.get(struct_key).append((tensor[Const.DTYPE], tensor[Const.SHAPE], tensor[Const.MD5]))
        else:
            op_dict.get(struct_key).append((tensor[Const.DTYPE], tensor[Const.SHAPE]))

        # 当统计量为None时，转成字符串None，避免后续操作list放到pd中时None被默认转成NaN
        op_dict[Const.SUMMARY].append(
            [str(tensor[key]) if tensor[key] is None else tensor[key] for key in Const.SUMMARY_METRICS_LIST])

        if dump_mode == Const.ALL:
            op_dict[Const.DATA_NAME].append(tensor.get(Const.DATA_NAME))

    if not op_dict[CompareConst.KWARGS_STRUCT]:
        del op_dict[CompareConst.KWARGS_STRUCT]
    return op_dict if op_dict[CompareConst.OP_NAME] else {}


def print_compare_ends_info():
    total_len = len(CompareConst.COMPARE_ENDS_SUCCESSFULLY) + Const.FILL_CHAR_NUMS
    logger.info('*' * total_len)
    logger.info(f"*{CompareConst.COMPARE_ENDS_SUCCESSFULLY.center(total_len - 2)}*")
    logger.info('*' * total_len)


def table_value_is_valid(value: str) -> bool:
    if not isinstance(value, str):
        return True
    try:
        # -1.00 or +1.00 should be considered as digit numbers
        float(value)
    except ValueError:
        # otherwise, they will be considered as formular injections
        return not bool(re.compile(FileCheckConst.CSV_BLACK_LIST).search(value))
    return True


class ApiBatch:
    def __init__(self, api_name: str, start: int):
        self.api_name = api_name
        self.start = start
        self.input_len = 1  # input的数量
        self.params_end_index = start + 1  # params的结束index
        self.output_end_index = start + 1  # output的结束index
        self.params_grad_end_index = start + 1  # params_grad的结束index
        # 内部state的标志("input", "output", "parameters", "parameters_grad"),
        # 用于控制计算input_len, output_end_index, params_end_index, self.params_grad_end_index
        self._state = Const.INPUT  # api_batch初始化为input

    def set_state(self, state: str):
        """设置当前状态"""
        if state in {Const.INPUT, Const.OUTPUT, Const.KWARGS, Const.PARAMS, Const.PARAMS_GRAD}:
            self._state = state
        else:
            raise ValueError(f"Invalid state: {state}")

    def increment(self, state: str):
        self.set_state(state)
        if self._state == Const.INPUT or self._state == Const.KWARGS:
            self.input_len += 1
            self.params_end_index += 1
            self.output_end_index += 1
        if self._state == Const.PARAMS:
            self.params_end_index += 1
            self.output_end_index += 1
        if self._state == Const.OUTPUT:
            self.output_end_index += 1
        self.params_grad_end_index += 1


def api_batches_update(api_batches, api_name, state, index):
    """
    当一个api的所有item更新完后，input, output的索引范围：
    input: [start: start+input_len]
    output: [start+input_len: output_end_index]
    params: [output_end_index: params_end_index]
    """
    if not api_batches:
        api_batches.append(ApiBatch(api_name, index))
    else:
        api_batch = api_batches[-1]
        if api_batch.api_name == api_name or (
                not re.search(Const.REGEX_FORWARD_BACKWARD, api_name) and api_name in api_batch.api_name):
            try:
                api_batch.increment(state)
            except ValueError as e:
                logger.error(f"api_batch: {api_batch} with invalid state, please check! {e}")
                raise CompareException(CompareException.INVALID_STATE_ERROR) from e
        else:
            api_batches.append(ApiBatch(api_name, index))


def reorder_index(op_parsed_list):
    """
    对单个api解析的op_items的index进行重排，将parameter的index放到output前面，返回新的重排后的index列表，op_parsed_list不变
    """
    index_param = []
    index_output = []
    index_param_grad = []
    index_other = []
    for i, op_item in enumerate(op_parsed_list[:-1]):
        state = op_item.get(Const.STATE)
        if state == Const.PARAMS:
            index_param.append(i)
        elif state == Const.OUTPUT:
            index_output.append(i)
        elif state == Const.PARAMS_GRAD:
            index_param_grad.append(i)
        else:
            index_other.append(i)
    # 合并others, parameters, 和output，确保parameters排在output前面
    reordered_index_list = index_other + index_param + index_output + index_param_grad
    return reordered_index_list


def reorder_op_name_list(op_name_list, state_list):
    if not op_name_list:
        return op_name_list, state_list

    parameters = []
    output = []
    parameters_grad = []
    others = []
    parameters_s = []
    output_s = []
    parameters_grad_s = []
    others_s = []
    for op_name, state in zip(op_name_list, state_list):
        if state == Const.PARAMS:
            parameters.append(op_name)
            parameters_s.append(state)
        elif state == Const.OUTPUT:
            output.append(op_name)
            output_s.append(state)
        elif state == Const.PARAMS_GRAD:
            parameters_grad.append(op_name)
            parameters_grad_s.append(state)
        else:
            others.append(op_name)
            others_s.append(state)
    # 合并others, parameters, 和output，确保parameters排在output前面
    op_name_reorder = others + parameters + output + parameters_grad
    state_reorder = others_s + parameters_s + output_s + parameters_grad_s
    return op_name_reorder, state_reorder


def process_summary_data(summary_data):
    """处理summary_data中的nan值，返回处理后的列表"""
    return [CompareConst.NAN if isinstance(x, float) and math.isnan(x) else x for x in summary_data]


def get_rela_diff_summary_mode(result_item, npu_summary_data, bench_summary_data, err_msg):
    start_idx = CompareConst.SUMMARY_COMPARE_RESULT_HEADER.index(CompareConst.MAX_DIFF)
    warning_flag = False
    for i, (npu_val, bench_val) in enumerate(zip(npu_summary_data, bench_summary_data)):
        if all(isinstance(val, (float, int)) and not isinstance(val, bool) for val in [npu_val, bench_val]):
            diff = npu_val - bench_val
            if math.isnan(diff):
                diff = CompareConst.NAN
                relative = CompareConst.NAN
            else:
                if bench_val != 0:
                    relative = str(abs((diff / bench_val) * 100)) + '%'
                else:
                    relative = CompareConst.N_A
                magnitude_diff = abs(diff) / (max(abs(npu_val), abs(bench_val)) + CompareConst.EPSILON)
                if magnitude_diff > CompareConst.MAGNITUDE:
                    warning_flag = True
            result_item[start_idx + i] = diff
            result_item[start_idx + i + CompareConst.STATISTICS_INDICATOR_NUM] = relative
        else:
            result_item[start_idx + i] = CompareConst.N_A
            result_item[start_idx + i + CompareConst.STATISTICS_INDICATOR_NUM] = CompareConst.N_A

    accuracy_check = CompareConst.WARNING if warning_flag else ""
    err_msg += "Need double check api accuracy." if warning_flag else ""
    for i in range(start_idx, len(result_item)):
        if str(result_item[i]) in ('inf', '-inf', 'nan'):
            result_item[i] = f'{result_item[i]}\t'
    return result_item, accuracy_check, err_msg


@dataclass
class ApiItemInfo:
    name: str
    struct: tuple
    stack_info: list


def stack_column_process(result_item, has_stack, index, key, npu_stack_info):
    if has_stack and index == 0 and key == CompareConst.INPUT_STRUCT:
        result_item.extend(npu_stack_info)
    else:
        result_item.append(CompareConst.NONE)
    return result_item


def result_item_init(n_info, b_info, requires_grad_pair, dump_mode):
    n_len = len(n_info.struct)
    b_len = len(b_info.struct)
    # requires_grad_pair内部创建，固定两个元素
    n_requires_grad = requires_grad_pair[0]
    b_requires_grad = requires_grad_pair[1]
    req_grad_consist = n_requires_grad == b_requires_grad
    struct_long_enough = (n_len > 2 and b_len > 2) if dump_mode == Const.MD5 else (n_len > 1 and b_len > 1)
    if struct_long_enough:
        result_item = [
            n_info.name, b_info.name, n_info.struct[0], b_info.struct[0], n_info.struct[1], b_info.struct[1],
            n_requires_grad, b_requires_grad
        ]
        if dump_mode == Const.MD5:
            md5_compare_result = CompareConst.PASS if n_info.struct[2] == b_info.struct[2] else CompareConst.DIFF
            result_item.extend([n_info.struct[2], b_info.struct[2], req_grad_consist, md5_compare_result])
        elif dump_mode == Const.SUMMARY:
            result_item.extend([" "] * 8)  # 8个统计量数据情况的比对指标
        else:
            result_item.extend([" "] * 6)  # 6个真实数据情况的比对指标
    else:
        err_msg = "index out of bounds error will occur in result_item_init, please check!\n" \
                  f"npu_info_struct is {n_info.struct}\n" \
                  f"bench_info_struct is {b_info.struct}"
        logger.error(err_msg)
        raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR)
    return result_item


def count_struct(op_dict):
    parts = [
        CompareConst.OP_NAME,
        CompareConst.INPUT_STRUCT,
        CompareConst.OUTPUT_STRUCT,
        CompareConst.PARAMS_STRUCT,
        CompareConst.PARAMS_GRAD_STRUCT
    ]
    lengths = [len(op_dict.get(part, [])) for part in parts]
    num = lengths[0]
    if num != sum(lengths[1:]):
        logger.error(f"Length of names and structs of op_dict not match. Please check! op_dict: {op_dict}")
        raise CompareException(CompareException.NAMES_STRUCTS_MATCH_ERROR)
    return tuple(lengths)


def get_accuracy(result, n_dict, b_dict, dump_mode):
    def get_accuracy_core(n_start, n_len, b_start, b_len, key):
        min_len = min(n_len, b_len)
        npu_stack_info = n_dict.get("stack_info", None)
        bench_stack_info = b_dict.get("stack_info", None)
        has_stack = npu_stack_info and bench_stack_info

        if dump_mode == Const.ALL:
            npu_data_name_list = n_dict.get("data_name", None)
            bench_data_name_list = b_dict.get("data_name", None)

        for index in range(min_len):
            n_name = safe_get_value(n_dict, n_start + index, "n_dict", key="op_name")
            b_name = safe_get_value(b_dict, b_start + index, "b_dict", key="op_name")
            n_struct = safe_get_value(n_dict, index, "n_dict", key=key)
            b_struct = safe_get_value(b_dict, index, "b_dict", key=key)
            n_requires_grad = safe_get_value(n_dict, n_start + index, "n_dict", key='requires_grad')
            b_requires_grad = safe_get_value(b_dict, b_start + index, "b_dict", key='requires_grad')
            requires_grad_pair = [n_requires_grad, b_requires_grad]
            req_grad_consist = n_requires_grad == b_requires_grad
            err_msg = ""

            npu_info = ApiItemInfo(n_name, n_struct, npu_stack_info)
            bench_info = ApiItemInfo(b_name, b_struct, bench_stack_info)
            result_item = result_item_init(npu_info, bench_info, requires_grad_pair, dump_mode)

            if dump_mode == Const.MD5:
                result_item = stack_column_process(result_item, has_stack, index, key, npu_stack_info)
                result.append(result_item)
                continue

            npu_summary_data = safe_get_value(n_dict, n_start + index, "n_dict", key=CompareConst.SUMMARY)
            bench_summary_data = safe_get_value(b_dict, b_start + index, "b_dict", key=CompareConst.SUMMARY)
            result_item.extend(process_summary_data(npu_summary_data))
            result_item.extend(process_summary_data(bench_summary_data))

            if dump_mode == Const.SUMMARY:
                result_item, accuracy_check, err_msg = get_rela_diff_summary_mode(result_item, npu_summary_data,
                                                                                  bench_summary_data, err_msg)

            result_item.append(req_grad_consist)
            err_msg += "Requires_grad inconsistent." if not req_grad_consist else ""
            result_item.append(accuracy_check if dump_mode == Const.SUMMARY else CompareConst.ACCURACY_CHECK_YES)
            result_item.append(err_msg)
            result_item = stack_column_process(result_item, has_stack, index, key, npu_stack_info)
            if dump_mode == Const.ALL:
                npu_data_name = safe_get_value(npu_data_name_list, n_start + index, "npu_data_name_list")
                bench_data_name = safe_get_value(bench_data_name_list, b_start + index, "bench_data_name_list")
                result_item.append([npu_data_name, bench_data_name])

            result.append(result_item)

        if n_len > b_len:
            for index in range(b_len, n_len):
                try:
                    n_name = safe_get_value(n_dict, n_start + index, "n_dict", key="op_name")
                    n_struct = safe_get_value(n_dict, index, "n_dict", key=key)
                    n_requires_grad = safe_get_value(n_dict, n_start + index, "n_dict", key='requires_grad')

                    if dump_mode == Const.MD5:
                        result_item = [
                            n_name, CompareConst.NAN, n_struct[0], CompareConst.NAN, n_struct[1], CompareConst.NAN,
                            n_requires_grad, CompareConst.NAN,
                            n_struct[2], CompareConst.NAN,
                            False,
                            CompareConst.NAN,
                            None
                        ]
                        result.append(result_item)
                        continue
                    if dump_mode == Const.ALL:
                        result_item = [
                            n_name, CompareConst.NAN, n_struct[0], CompareConst.NAN, n_struct[1], CompareConst.NAN,
                            n_requires_grad, CompareConst.NAN,
                            " ", " ", " ", " ", " ", " "
                        ]
                    else:
                        result_item = [
                            n_name, CompareConst.NAN, n_struct[0], CompareConst.NAN, n_struct[1], CompareConst.NAN,
                            n_requires_grad, CompareConst.NAN,
                            " ", " ", " ", " ", " ", " ", " ", " "
                        ]
                    summary_data = n_dict.get(CompareConst.SUMMARY)[n_start + index]
                    result_item.extend(summary_data)
                    summary_data = [CompareConst.NAN for _ in range(len(n_dict.get(CompareConst.SUMMARY)[0]))]
                    result_item.extend(summary_data)
                    result_item.append(False)
                except IndexError as e:
                    err_msg = "index out of bounds error occurs, please check!\n" \
                              f"n_dict is {n_dict}"
                    logger.error(err_msg)
                    raise CompareException(CompareException.INDEX_OUT_OF_BOUNDS_ERROR) from e

                err_msg = ""
                result_item.append(CompareConst.PASS)
                result_item.append(err_msg)
                result_item = stack_column_process(result_item, has_stack, index, key, npu_stack_info)
                if dump_mode == Const.ALL:
                    npu_data_name = safe_get_value(npu_data_name_list, n_start + index, "npu_data_name_list")
                    result_item.append([npu_data_name, "-1"])

                result.append(result_item)

    _, n_num_input, n_num_output, n_num_params, n_num_params_grad = count_struct(n_dict)
    _, b_num_input, b_num_output, b_num_params, b_num_params_grad = count_struct(b_dict)

    get_accuracy_core(0, n_num_input, 0, b_num_input, CompareConst.INPUT_STRUCT)
    get_accuracy_core(n_num_input, n_num_params, b_num_input, b_num_params,
                      CompareConst.PARAMS_STRUCT)
    get_accuracy_core(n_num_input + n_num_params, n_num_output, b_num_input + b_num_params, b_num_output,
                      CompareConst.OUTPUT_STRUCT)
    get_accuracy_core(n_num_input + n_num_params + n_num_output, n_num_params_grad,
                      b_num_input + b_num_params + b_num_output, b_num_params_grad,
                      CompareConst.PARAMS_GRAD_STRUCT)


def make_result_table(result, dump_mode, stack_mode):
    header = CompareConst.HEAD_OF_COMPARE_MODE[dump_mode][:]

    if stack_mode:
        header.append(CompareConst.STACK)
        if dump_mode == Const.ALL:
            header.append(CompareConst.DATA_NAME)
    else:
        if dump_mode == Const.ALL:
            for row in result:
                del row[-2]  # 输出结果不要堆栈信息时，删除中间结果result中的stack info，真实数据时为倒数第2列
            header.append(CompareConst.DATA_NAME)
        else:
            for row in result:
                del row[-1]  # 输出结果不要堆栈信息时，删除中间结果result中的stack info，非真实数据时为倒数第1列
    result_df = pd.DataFrame(result, columns=header, dtype='object')
    return result_df


def gen_api_batches(result: np.ndarray, header: list):
    api_name_index = header.index(Const.API_ORIGIN_NAME)
    state_name_index = header.index(Const.STATE)
    api_batches = []
    for i, res_i in enumerate(result):
        api_name = safe_get_value(res_i, api_name_index, "res_i")
        state = safe_get_value(res_i, state_name_index, "res_i")
        api_batches_update(api_batches, api_name, state, i)
    return api_batches


def get_paired_dirs(npu_path, bench_path):
    npu_dirs = set(os.listdir(npu_path))
    bench_dirs = set(os.listdir(bench_path))
    return list(npu_dirs & bench_dirs)


def _compare_parser(parser):
    parser.add_argument("-m", "--mode", dest="mode", type=str, default="auto",
                        help="<optional> Comparison mode: 'auto' (default) or 'torchair' for torchair accuracy compare",
                        required=False)
    parser.add_argument("-tp", "--target_path", dest="target_path", type=str,
                        help="<Required> The compare target device path", required=True)
    parser.add_argument("-gp", "--golden_path", dest="golden_path", type=str,
                        help="<Required> The compare golden device path", required=True)
    parser.add_argument("-o", "--output_path", dest="output_path", type=str,
                        help="<Required> The compare task result out path. Default path: ./output",
                        required=False, default="./output", nargs="?", const="./output")
    parser.add_argument("-fm", "--fuzzy_match", dest="fuzzy_match", action="store_true",
                        help="<optional> Whether to perform a fuzzy match on the api name.", required=False)
    parser.add_argument("-cm", "--cell_mapping", dest="cell_mapping", type=str, nargs='?', const=True,
                        help="<optional> The cell mapping file path.", required=False)
    parser.add_argument("-am", "--api_mapping", dest="api_mapping", type=str, nargs='?', const=True,
                        help="<optional> The api mapping file path.", required=False)
    parser.add_argument("-dm", "--data_mapping", dest="data_mapping", type=str,
                        help="<optional> The data mapping file path.", required=False)
    parser.add_argument("-lm", "--layer_mapping", dest="layer_mapping", type=str, nargs='?', const=True,
                        help="<optional> The layer mapping file path.", required=False)
    parser.add_argument("-da", "--diff_analyze", dest="diff_analyze", action="store_true",
                        help="<optional> Whether to perform a diff analyze on the api name.", required=False)
    parser.add_argument("-tensor_log", "--is_print_compare_log", dest="is_print_compare_log", action="store_true",
                        help="<Optional> whether print compare log for compare auto mode task.", required=False)
    parser.add_argument("-fr", "--fusion_rule_file", dest="fusion_rule_file", type=str, default="",
                        help="<Optional> The fusion rule file path.", required=False)
    parser.add_argument("-qfr", "--quant_fusion_rule_file", dest="quant_fusion_rule_file", type=str, default="",
                        help="<Optional> The quant fusion rule file path.", required=False)
    parser.add_argument("-cfr", "--close_fusion_rule_file", dest="close_fusion_rule_file", type=str, default="",
                        help="<Optional> The close fusion rule file path.", required=False)
    # rank：ms静态图比对、推理离线模型一键式比对
    parser.add_argument("--rank", dest="rank", type=str, required=False,
                        help="<optional> Ranks to compare when compare kernel of MindSpore for <compare auto>. "
                             "Ranks to compare for <torchair dumps>. "
                             "Input rank ID [0, 255] for <compare offline_model>.")
    parser.add_argument("--step", dest="step", type=str, required=False,
                        help="<optional> Steps to compare when compare kernel of MindSpore.")
    parser.add_argument('--input_data', dest="input_data", default='',
                        help='The input data path of the model. Separate multiple inputs with commas(,).'
                             ' E.g: input_0.bin,input_1.bin')
    parser.add_argument('--input_shape', dest="input_shape", default='',
                        help="Shape of input shape. Separate multiple nodes with semicolons(;)."
                             " E.g: \"input_name1:1,224,224,3;input_name2:3,300\"")
    parser.add_argument('--output_size', dest="output_size", default='',
                        help='The size of output. Separate multiple sizes with commas(,). E.g: 10200,34000')
    parser.add_argument('--dym_shape_range', dest="dym_shape_range", default='',
                        help="Dynamic shape range using in dynamic model, using this means ignore input_shape."
                             " E.g: \"input_name1:1,3,200\~224,224-230;input_name2:1,300\"")
    parser.add_argument('-ofs', '--onnx_fusion_switch', dest="onnx_fusion_switch", default=True,
                        help='Onnxruntime fusion switch, set False for dump complete onnx data when necessary. '
                             'Usage: -ofs False')


def get_sorted_ranks(npu_dump_dir, bench_dump_dir):
    """
    get the ranks and match by order
    """ 
    unsorted_npu_ranks = check_and_return_dir_contents(npu_dump_dir, 'rank')
    unsorted_bench_ranks = check_and_return_dir_contents(bench_dump_dir, 'rank')
    # 正则匹配已经校验rank后面必是数字，或者无数字的rank
    npu_ranks = sorted(unsorted_npu_ranks, key=lambda x: int(x[4:]) if len(x) > 4 else -1)  # 前四个字符都是rank，后面是卡号
    bench_ranks = sorted(unsorted_bench_ranks, key=lambda x: int(x[4:]) if len(x) > 4 else -1)
    if len(npu_ranks) != len(bench_ranks):
        logger.error('The number of ranks in the two runs are different. '
                     'Unable to match the ranks. Please use another folder to compare '
                     'or use compare() api and manually match the ranks.')
        raise CompareException(CompareException.INVALID_PATH_ERROR)
    return npu_ranks, bench_ranks


def multi_statistics_compare(func, func_args):
    def err_call(args):
        logger.error(f'Multiprocess statistics compare failed! Reason: {args}')

    compare_func, input_param_nr_list, output_path, kwargs = func_args

    param_num = len(input_param_nr_list)
    process_num = max(int((multiprocessing.cpu_count() + 1) // 4), 1)
    if param_num <= process_num:
        process_num = param_num
        chunks = [[input_param_nr] for input_param_nr in input_param_nr_list]
    else:
        chunk_size = param_num // process_num
        remainder = param_num % process_num
        chunks = [input_param_nr_list[i:i + chunk_size] for i in range(0, param_num - remainder, chunk_size)]
        for i in range(remainder):
            chunks[i].append(input_param_nr_list[param_num - remainder + i])

    pool = multiprocessing.Pool(process_num)

    async_results = []
    for chunk in chunks:
        result = pool.apply_async(func, args=(compare_func, chunk, output_path, kwargs), error_callback=err_call)
        async_results.append(result)

    pool.close()

    for ar in async_results:
        try:
            ar.get(timeout=3600)
        except Exception as e:
            logger.error(f"Task failed with exception: {e}")
            pool.terminate()
            raise CompareException(CompareException.MULTIPROCESS_ERROR) from e

    pool.join()


def mp_logger_init(ranks_str):
    """
    多进程比对需要对logger进行wrap和patch，在日志前加上卡号信息，从而实现不同进程日志的隔离
    """

    def wrap_logger(fn):
        def inner(msg, *args, **kwargs):
            return fn(ranks_str + msg, *args, **kwargs)

        return inner

    logger.info = wrap_logger(logger.info)
    logger.warning = wrap_logger(logger.warning)
    logger.error = wrap_logger(logger.error)


def multi_ranks_compare(compare_func, input_param_nr_list, output_path, kwargs):
    """
    将多卡数据分成多进程后，单进程内可能还有多张卡的数据，因此还需要多次比对
    """
    rank_list = [input_param_nr[1] for input_param_nr in input_param_nr_list]  # input_param_nr内部数据结构，2元素tuple
    ranks_str = f"[{' '.join(rank_list)}]"
    mp_logger_init(ranks_str)
    for input_param_nr in input_param_nr_list:
        input_param, nr = input_param_nr
        compare_entry(compare_func, input_param, output_path, nr, kwargs)


def compare_entry(compare_func, input_param, output_path, nr, kwargs):
    try:
        compare_func(input_param=input_param, output_path=output_path, suffix=f'_{nr}', **kwargs)
    except CompareException as e:
        if e.code == CompareException.INVALID_DATA_ERROR:
            logger.error(f"Invalid or missing 'data' in dump.json. Skipping {nr} comparison.")
        if e.code == CompareException.INVALID_TASK_ERROR:
            logger.error(f"Invalid or missing 'task' in dump.json. Skipping {nr} comparison.")


def compare_distributed_inner(npu_dump_dir, bench_dump_dir, output_path, **kwargs):
    def extract_compare_param(_file_type):
        npu_data_dir = os.path.join(npu_dump_dir, nr)
        bench_data_dir = os.path.join(bench_dump_dir, br)
        npu_path = extract_json(npu_data_dir, _file_type)
        bench_path = extract_json(bench_data_dir, _file_type)
        if npu_path == "" or bench_path == "":
            logger.debug(f'Did not find paired {_file_type} in {nr} and {br}, skip comparing.')
            return {}, True
        _input_param = {
            'npu_path': npu_path,
            'bench_path': bench_path,
            'is_print_compare_log': kwargs.get('is_print_compare_log', False)
        }
        return _input_param, False

    if kwargs.get('suffix'):
        logger.error("Argument 'suffix' is not supported for compare_distributed.")
        raise CompareException(CompareException.INVALID_PARAM_ERROR)

    npu_ranks, bench_ranks = get_sorted_ranks(npu_dump_dir, bench_dump_dir)

    # ------------------预载rank0的json用于判断是什么类型dump数据------------------
    # 判断是否存在dump.json或debug.json
    if npu_ranks:
        dir_path = os.path.join(npu_dump_dir, npu_ranks[0])
        dump_file = os.path.join(dir_path, 'dump.json')
        debug_file = os.path.join(dir_path, 'debug.json')

        # 确定pre_check_dump_path
        if os.path.exists(dump_file):
            pre_check_dump_path = dump_file
        elif os.path.exists(debug_file):
            pre_check_dump_path = debug_file
        else:
            pre_check_dump_path = ''
    else:
        pre_check_dump_path = ''

    # 如果pre_check_dump_path为空，直接返回
    if not pre_check_dump_path:
        return

    dump_data = load_json(pre_check_dump_path)

    compare_framework = dump_data.get("framework", None)
    if compare_framework == Const.PT_FRAMEWORK:
        from msprobe.pytorch.compare.pt_compare import pt_compare
        compare_func = pt_compare
    elif compare_framework == Const.MS_FRAMEWORK:
        from msprobe.mindspore.compare.ms_compare import ms_compare
        compare_func = ms_compare
    else:
        logger.error(f"Unrecognized framework, now is {compare_framework}, please check dump.json.")
        raise CompareException(CompareException.INVALID_TASK_ERROR)

    compare_json_file_type = Const.DUMP_JSON_FILE
    if dump_data.get('level') == Const.LEVEL_DEBUG:
        compare_json_file_type = Const.DEBUG_JSON_FILE
    # ------------------统计量、md5比对------------------
    if dump_data.get('task') == Const.STATISTICS:
        # dump数据为统计量或md5时，多进程加速比对
        input_param_nr_list = []
        for nr, br in zip(npu_ranks, bench_ranks):
            input_param, skip = extract_compare_param(compare_json_file_type)
            if not skip:
                input_param_nr_list.append((input_param, nr))
        func_args = (compare_func, input_param_nr_list, output_path, kwargs)
        multi_statistics_compare(multi_ranks_compare, func_args)
        return

    # ------------------真实数据比对------------------
    for nr, br in zip(npu_ranks, bench_ranks):
        input_param, skip = extract_compare_param(compare_json_file_type)
        if not skip:
            compare_entry(compare_func, input_param, output_path, nr, kwargs)


def check_input_param_path(input_param):
    npu_path = input_param.get("npu_path", None)
    bench_path = input_param.get("bench_path", None)
    check_file_or_directory_path(npu_path)
    check_file_or_directory_path(bench_path)


def get_compare_framework(target_path, golden_path):
    target_dump_json_content = load_json(target_path)
    golden_dump_json_content = load_json(golden_path)
    target_framework = target_dump_json_content.get("framework", None)
    golden_framework = golden_dump_json_content.get("framework", None)
    if target_framework == Const.PT_FRAMEWORK and golden_framework == Const.PT_FRAMEWORK:
        frame_name = Const.PT_FRAMEWORK
    elif (target_framework in [Const.MS_FRAMEWORK, Const.MT_FRAMEWORK] and
          golden_framework in [Const.PT_FRAMEWORK, Const.MS_FRAMEWORK]):
        frame_name = Const.MS_FRAMEWORK
    else:
        logger.error(f"Unrecognized framework, target now is {target_framework}, golden now is {golden_framework},"
                     f" please check dump.json.")
        raise CompareException(CompareException.INVALID_TASK_ERROR)
    return frame_name


def check_input_param_path_and_framework(args, target_framework):
    compare_framework = get_compare_framework(args.target_path, args.golden_path)
    if compare_framework != target_framework:
        logger.error(f"Expected frame to be {target_framework}, actual is {compare_framework}, please check dump.json.")
        raise CompareException(CompareException.INVALID_TASK_ERROR)

    if target_framework == Const.PT_FRAMEWORK and not is_module_available("torch"):
        logger.error("PyTorch does not exist, please install PyTorch library")
        raise Exception("PyTorch does not exist, please install PyTorch library")
    if target_framework == Const.MS_FRAMEWORK and not is_module_available("mindspore"):
        logger.error("MindSpore does not exist, please install MindSpore library")
        raise Exception("MindSpore does not exist, please install MindSpore library")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import random
import stat
import inspect
from functools import wraps

import numpy as np
import torch
import torch.distributed as dist

from msprobe.core.common.exceptions import DistributedNotInitializedError
from msprobe.core.common.file_utils import (FileCheckConst, change_mode,
                                            check_file_or_directory_path, check_path_before_create, FileOpen)
from msprobe.core.common.log import logger
from msprobe.core.common.utils import check_seed_all, is_save_variable_valid
from packaging import version

try:
    import torch_npu
except ImportError:
    is_gpu = True
else:
    is_gpu = False

torch_without_guard_version = torch.__version__ >= '2.1'
torch_version_above_or_equal_2 = torch.__version__.split('+')[0] >= '2.0'

if not is_gpu and not torch_without_guard_version:
    from torch_npu.utils.device_guard import torch_device_guard as torch_npu_device_guard

npu_distributed_api = ['isend', 'irecv']


def parameter_adapter(func):
    def handle_masked_select(input_tensor, indices):
        masked_select_func = getattr(torch._C._VariableFunctionsClass, "masked_select")
        if input_tensor.dtype == torch.bfloat16:
            # masked_select在NPU上输入数据dtype类型为bfloat16会报错，提示不支持此类型
            return masked_select_func(input_tensor.to(torch.float32), indices).to(torch.bfloat16)
        else:
            return masked_select_func(input_tensor, indices)

    @wraps(func)
    def inner(self, *args, **kwargs):
        if self.api_name == "__getitem__" and len(args) > 1 and isinstance(args[1], torch.Tensor):
            input_tensor = args[0]
            indices = args[1]
            if indices.dtype == torch.uint8:
                indices = indices.bool()
            if indices.dtype == torch.bool:
                if indices.shape == input_tensor.shape:
                    return handle_masked_select(input_tensor, indices)
                else:
                    indices = getattr(torch._C._VariableFunctionsClass, "nonzero")(indices, as_tuple=True)
                    return getattr(torch._C._TensorBase, "__getitem__")(input_tensor, indices)
            elif indices.dtype != torch.bool:
                if not indices.shape or len(indices.shape) == 1:
                    return func(self, input_tensor, indices.tolist())
                elif len(indices.shape) == 2:
                    result = [func(self, input_tensor, index) for index in indices.tolist()]
                    return getattr(torch._C._VariableFunctionsClass, "stack")(result, 0)
                else:
                    res = [input_tensor[tensor_index] for tensor_index in indices]
                    return getattr(torch._C._VariableFunctionsClass, "stack")(res, 0)
        if self.api_name == "__eq__" and len(args) > 1 and args[1] is None:
            return False
        return func(self, *args, **kwargs)

    return inner


def torch_device_guard(func):
    if is_gpu or torch_without_guard_version:
        return func

    # Parse args/kwargs matched torch.device objects
    @torch_npu_device_guard
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def get_rank_if_initialized():
    """
        return rank id if it is initialized or raise Exception: DistributedNotInitializedError
    """
    if torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    else:
        raise DistributedNotInitializedError("torch distributed environment is not initialized")


def remove_dropout():
    if torch.__version__ > "1.8":
        logger.info_on_rank_0("For precision comparison, the probability p in the dropout method is set to 0.")
        import torch.nn.functional as F
        from torch import _VF
        from torch.overrides import has_torch_function_unary, handle_torch_function

        def function_dropout(input_tensor: torch.Tensor, p: float = 0.5, training: bool = True,
                             inplace: bool = False) -> torch.Tensor:
            if has_torch_function_unary(input_tensor):
                return handle_torch_function(
                    function_dropout, (input_tensor,), input_tensor, p=0., training=training, inplace=inplace)
            if p < 0.0 or p > 1.0:
                raise ValueError("dropout probability has to be between 0 and 1, " "but got {}".format(p))
            return _VF.dropout_(input_tensor, 0., training) if inplace else _VF.dropout(input_tensor, 0., training)

        def function_dropout2d(input_tensor: torch.Tensor, p: float = 0.5, training: bool = True,
                               inplace: bool = False) -> torch.Tensor:
            if has_torch_function_unary(input_tensor):
                return handle_torch_function(
                    function_dropout2d, (input_tensor,), input_tensor, p=0., training=training, inplace=inplace)
            if p < 0.0 or p > 1.0:
                raise ValueError("dropout probability has to be between 0 and 1, " "but got {}".format(p))
            return _VF.feature_dropout_(input_tensor, 0., training) if inplace else _VF.feature_dropout(input_tensor,
                                                                                                        0., training)

        def function_dropout3d(input_tensor: torch.Tensor, p: float = 0.5, training: bool = True,
                               inplace: bool = False) -> torch.Tensor:
            if has_torch_function_unary(input_tensor):
                return handle_torch_function(
                    function_dropout3d, (input_tensor,), input_tensor, p=0., training=training, inplace=inplace)
            if p < 0.0 or p > 1.0:
                raise ValueError("dropout probability has to be between 0 and 1, " "but got {}".format(p))
            return _VF.feature_dropout_(input_tensor, 0., training) if inplace else _VF.feature_dropout(input_tensor,
                                                                                                        0., training)

        F.dropout = function_dropout
        F.dropout2d = function_dropout2d
        F.dropout3d = function_dropout3d


def seed_all(seed=1234, mode=False, rm_dropout=False):
    check_seed_all(seed, mode, rm_dropout)
    try:
        random.seed(seed)
        os.environ['PYTHONHASHSEED'] = str(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        cuda_version = torch.version.cuda
        if cuda_version is not None and version.parse(cuda_version) >= version.parse("10.2"):
            os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'
        os.environ['HCCL_DETERMINISTIC'] = str(mode)
        torch.use_deterministic_algorithms(mode)
        if is_gpu:
            torch.cuda.manual_seed_all(seed)
            torch.cuda.manual_seed(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.enable = False
            torch.backends.cudnn.benchmark = False
        else:
            torch_npu.npu.manual_seed_all(seed)
            torch_npu.npu.manual_seed(seed)
        if rm_dropout:
            remove_dropout()
    except Exception as e:
        logger.error(f"There is an unexpected error while determinating randomness. {e}")


class Const:
    """
    Class for const
    """
    SEP = "."
    MODEL_TYPE = ['.onnx', '.pb', '.om']
    DIM_PATTERN = r"^(-?[0-9]+)(,-?[0-9]+)*"
    SEMICOLON = ";"
    COLON = ":"
    EQUAL = "="
    COMMA = ","
    DOT = "."
    DUMP_RATIO_MAX = 100
    SUMMERY_DATA_NUMS = 256
    FLOAT_EPSILON = np.finfo(float).eps
    SUPPORT_DUMP_MODE = ['api', 'acl']
    ON = 'ON'
    OFF = 'OFF'
    KWARGS = 'kwargs'
    INPUT = 'input'
    OUTPUT = 'output'
    BACKWARD = 'backward'
    FORWARD = 'forward'
    PRE_FORWARD = "pre_forward"
    INPUT_ARGS = 'input_args'
    INPUT_KWARGS = 'input_kwargs'
    GRAD_INPUT = 'grad_input'
    GRAD_OUTPUT = 'grad_output'
    START = "start"
    STOP = "stop"
    MAX = 'Max'
    MIN = 'Min'

    # dump mode
    ALL = "all"
    LIST = "list"
    RANGE = "range"
    STACK = "stack"
    ACL = "acl"
    API_LIST = "api_list"
    API_STACK = "api_stack"
    DUMP_MODE = [ALL, LIST, RANGE, STACK, ACL, API_LIST, API_STACK]
    AUTO = "auto"
    ONLINE_DUMP_MODE = [ALL, LIST, AUTO, OFF]
    SUMMARY = "summary"
    MD5 = "md5"
    SUMMARY_MODE = [ALL, SUMMARY, MD5]

    WRITE_FLAGS = os.O_WRONLY | os.O_CREAT
    OVERWRITE_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR

    PKL_SUFFIX = ".pkl"
    NUMPY_SUFFIX = ".npy"
    ONE_GB = 1 * 1024 * 1024 * 1024
    TEN_GB = 10 * 1024 * 1024 * 1024
    FILE_PATTERN = r'^[a-zA-Z0-9_./-]+$'
    FILE_NAME_LENGTH = 255
    DIRECTORY_LENGTH = 4096
    DISTRIBUTED_PREFIX_LENGTH = 60
    SUMMARY_COLUMN_NUM = 6
    STACK_COLUMN_NUM = 2
    # env dump path
    ASCEND_WORK_PATH = "ASCEND_WORK_PATH"
    DUMP_DIR = "dump_data"
    DATA = "data"

    ENV_ENABLE = "1"
    ENV_DISABLE = "0"

    MAX_SEED_VALUE = 2 ** 32 - 1

    TASK_LIST = ["tensor", "statistics", "overflow_check", "free_benchmark"]
    LEVEL_LIST = ["L0", "L1", "L2", "mix"]
    STATISTICS = "statistics"
    TENSOR = "tensor"
    OVERFLOW_CHECK = "overflow_check"
    FREE_BENCHMARK = "free_benchmark"

    ATTR_NAME_PREFIX = "wrap_"

    FLOAT_TYPE = [np.half, np.single, float, np.double, np.float64, np.longdouble, np.float32, np.float16]
    BOOL_TYPE = [bool, np.uint8]
    INT_TYPE = [np.int32, np.int64]
    NPU = 'NPU'
    DISTRIBUTED = 'Distributed'

    HIFLOAT8_TYPE = "torch_npu.HiFloat8Tensor"
    FLOAT8_E5M2_TYPE = "torch.float8_e5m2"
    FLOAT8_E4M3FN_TYPE = "torch.float8_e4m3fn"

    RAISE_PRECISION = {
        torch.float16: torch.float32,
        torch.bfloat16: torch.float32,
        torch.float32: torch.float64
    }
    CONVERT = {
        "int32_to_int64": ["torch.int32", "torch.int64"],
    }

    CONVERT_API = {
        "int32_to_int64": ["cross_entropy"]
    }

    DROPOUT_API_LIST = [
        "Functional.dropout.",
        "Functional.dropout1d.",
        "Functional.dropout2d.",
        "Functional.dropout3d.",
        "Functional.alpha_dropout.",
        "Functional.feature_alpha_dropout.",
        "Torch.dropout.",
        "Torch.dropout_.",
        "Torch.alpha_dropout.",
        "Torch.alpha_dropout_.",
        "Torch.feature_alpha_dropout.",
        "Torch.feature_alpha_dropout_.",
        "Torch.feature_dropout.",
        "Torch.feature_dropout_.",
        "NPU._npu_dropout."
    ]


def get_tensor_rank(in_feat, out_feat):
    if dist.is_initialized():
        return dist.get_rank()

    def get_tensor_rank_single(x):
        if isinstance(x, (list, tuple)):
            if len(x) > 0:
                return get_tensor_rank_single(x[0])
        elif isinstance(x, torch.Tensor):
            device = x.device
            if device.type != 'cpu':
                return device.index
        return None

    in_rank = get_tensor_rank_single(in_feat)
    out_rank = get_tensor_rank_single(out_feat)
    tensor_rank = in_rank if in_rank else out_rank
    return tensor_rank


def get_rank_id():
    if torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    return 0


def print_rank_0(message):
    if dist.is_initialized():
        if dist.get_rank() == 0:
            logger.info(message)
    else:
        logger.info(message)


def load_pt(pt_path, to_cpu=False, weights_only=True):
    pt_path = os.path.realpath(pt_path)
    check_file_or_directory_path(pt_path)
    try:
        if to_cpu:
            pt = torch.load(pt_path, map_location=torch.device("cpu"), weights_only=weights_only)
        else:
            pt = torch.load(pt_path, weights_only=weights_only)
    except Exception as e:
        raise RuntimeError(f"load pt file {pt_path} failed") from e
    return pt


def save_pt(tensor, filepath):
    check_path_before_create(filepath)
    filepath = os.path.realpath(filepath)
    try:
        torch.save(tensor, filepath)
    except Exception as e:
        logger.error("Save pt file failed, please check according possible error causes: "
                     "1. out of disk space or disk error, "
                     "2. no permission to write files, etc.")
        raise RuntimeError(f"save pt file {filepath} failed") from e
    change_mode(filepath, FileCheckConst.DATA_FILE_AUTHORITY)


def is_recomputation(call_stack=None):
    """Check if the current operation is in the re-computation phase.

    This function inspects the current call stack to indicate whether the current operation is in the
    re-computation phase. We use a blacklist mechanism, now supported megatron and mindspeed framework.
    megatron: The 'backward' function is called by the 'torch/autograd/function.py' file.
    mindspeed: The 'checkpoint_function_backward' function is called by the 'torch/autograd/function.py'
    file or the custom module(use CheckpointWithoutOutput) with the 'recompute_fn' function is executed within the
    'torch/utils/checkpoint.py' file.

    Returns:
        bool: True if in the re-computation phase, False otherwise.
    """
    if not call_stack:
        try:
            call_stack = inspect.stack()
        except Exception as e:
            logger.warning(
                f"Failed to capture stack trace, recomputation validation may be incorrect, error info: {e}.")
            return False
    backward_function_indices = []

    # Identify the function 'backward' is being executed within the 'torch/_tensor.py' file.
    for frame_info in call_stack:
        if frame_info.function == "recompute_fn" and frame_info.filename.endswith('torch/utils/checkpoint.py'):
            del call_stack
            return True

    # Identify indices in the call stack where the specific function is being executed
    for idx, frame_info in enumerate(call_stack):
        if (frame_info.function == Const.BACKWARD or frame_info.function == 'checkpoint_function_backward' and
                "megatron" in frame_info.filename):
            backward_function_indices.append(idx)

    # Check if the execution is within 'torch/autograd/function.py' file
    for idx in backward_function_indices:
        # The Megatron and MindSpeed L0&L1 scenes
        if idx + 1 < len(call_stack) and call_stack[idx + 1].filename.endswith('torch/autograd/function.py'):
            del call_stack
            return True
        # The latest MindSpeed L2 and ModelLink scenes
        if idx + 2 < len(call_stack) and call_stack[idx + 2].filename.endswith('torch/autograd/function.py'):
            del call_stack
            return True

    del call_stack
    return False


def check_save_param(variable, name, save_backward):
    # try catch this api to skip invalid call
    valid_data_types = (torch.Tensor, int, float, str)
    if not is_save_variable_valid(variable, valid_data_types):
        valid_data_types_with_nested_types = valid_data_types + (dict, tuple, list)
        logger.warning("PrecisionDebugger.save variable type not valid, "
                       f"should be one of {valid_data_types_with_nested_types}"
                       "Skip current save process.")
        raise ValueError
    if not isinstance(name, str):
        logger.warning("PrecisionDebugger.save name not valid, "
                       "should be string. "
                       "skip current save process.")
        raise ValueError
    if not isinstance(save_backward, bool):
        logger.warning("PrecisionDebugger.save_backward name not valid, "
                       "should be bool. "
                       "Skip current save process.")
        raise ValueError


def is_torch_nn_module(variable):
    return isinstance(variable, torch.nn.Module) and not isinstance(variable, torch.jit.ScriptModule)


def register_forward_pre_hook(module, forward_pre_hook):
    if torch_version_above_or_equal_2:
        module.register_forward_pre_hook(forward_pre_hook, with_kwargs=True)
    else:
        module.register_forward_pre_hook(forward_pre_hook)


def register_forward_hook(module, forward_hook):
    if torch_version_above_or_equal_2:
        module.register_forward_hook(forward_hook, with_kwargs=True)
    else:
        module.register_forward_hook(forward_hook)


def is_hifloat8_tensor(tensor):
    if not is_gpu and hasattr(torch_npu, "HiFloat8Tensor") and isinstance(tensor, torch_npu.HiFloat8Tensor):
        return True
    return False


def is_float8_tensor(tensor):
    return str(tensor.dtype) in [Const.FLOAT8_E5M2_TYPE, Const.FLOAT8_E4M3FN_TYPE] or is_hifloat8_tensor(tensor)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (c) 2025, Huawei Technologies.
# Adapt to the model hierarchical visualization data collected by the msprobe tool
# ==============================================================================

"""The TensorBoard Graphs plugin."""

import os
from tensorboard.plugins import base_plugin
from tensorboard.util import tb_logging

from .app.views.graph_views import GraphView
from .app.utils.graph_utils import GraphUtils
from .app.utils.global_state import GraphState
from .app.utils.constant import Extension

logger = tb_logging.get_logger()

PLUGIN_NAME = 'graph_ascend'
PLUGIN_NAME_RUN_METADATA_WITH_GRAPH = 'graph_ascend_run_metadata_graph'
DB_EXT = Extension.DB.value
JSON_EXT = Extension.JSON.value


class GraphsPlugin(base_plugin.TBPlugin):
    """Graphs Plugin for TensorBoard."""

    plugin_name = PLUGIN_NAME

    def __init__(self, context):
        """Instantiates GraphsPlugin via TensorBoard core.

        Args:
          context: A base_plugin.TBContext instance.
        """
        super().__init__(context)
        GraphState.reset_global_state()
        self.logdir = os.path.abspath(os.path.expanduser(context.logdir.rstrip('/')))
        # 将logdir赋值给global_state中的logdir属性,方便其他模块使用
        GraphState.set_global_value('logdir', self.logdir)

    def get_plugin_apps(self):
        return {     
            '/index.js': GraphView.static_file_route,
            '/index.html': GraphView.static_file_route,
            "/load_meta_dir": GraphView.load_meta_dir,
            "/screen": GraphView.search_node,
            '/loadGraphData': GraphView.load_graph_data,
            '/loadGraphConfigInfo': GraphView.load_graph_config_info,
            '/loadGraphAllNodeList': GraphView.load_graph_all_node_list,
            '/changeNodeExpandState': GraphView.change_node_expand_state,
            '/updateHierarchyData': GraphView.update_hierarchy_data,
            '/getNodeInfo': GraphView.get_node_info,
            '/addMatchNodes': GraphView.add_match_nodes,
            '/addMatchNodesByConfig': GraphView.add_match_nodes_by_config,
            '/deleteMatchNodes': GraphView.delete_match_nodes,
            '/saveData': GraphView.save_data,
            '/updateColors': GraphView.update_colors,
            '/saveMatchedRelations': GraphView.save_matched_relations,
            '/updatePrecisionError': GraphView.update_precision_error,
        }

    def is_active(self):
        """The graphs plugin is active if any run has a graph."""

        def _is_vis(path, file_name):
            return os.path.isfile(path) and (file_name.endswith(DB_EXT) or file_name.endswith(JSON_EXT))
        
        _, error = GraphUtils.safe_check_load_file_path(self.logdir, True)
        if error:
            return False
        for content in os.listdir(self.logdir):
            content_path = os.path.join(self.logdir, content)
            if _is_vis(content_path, content):
                return True
            if os.path.isdir(content_path):
                _, error = GraphUtils.safe_check_load_file_path(content_path, True)
                if error:
                    continue
                for file in os.listdir(content_path):
                    if _is_vis(os.path.join(content_path, file), file):
                        return True
        return False

    def data_plugin_names(self):
        return (
            PLUGIN_NAME,
            PLUGIN_NAME_RUN_METADATA_WITH_GRAPH,
        )

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            es_module_path='/index.js',
            disable_reload=True,
        )



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from tensorboard.util import tb_logging

from .graph_service_base import GraphServiceStrategy
from ..repositories.graph_repo_db import GraphRepoDB
from ..utils.global_state import GraphState
from ..utils.graph_utils import GraphUtils

from ..model.layout_hierarchy_model import LayoutHierarchyModel
from ..model.match_nodes_model import MatchNodesController
from ..utils.constant import NPU, BENCH, SINGLE, UN_MATCHED_VALUE
from ..utils.constant import MAX_RELATIVE_ERR, MIN_RELATIVE_ERR, MEAN_RELATIVE_ERR, NORM_RELATIVE_ERR

logger = tb_logging.get_logger()


class DbGraphService(GraphServiceStrategy):

    def __init__(self, run_path, tag):
        super().__init__(run_path, tag)
        runs = GraphState.get_global_value('runs')
        db_path = os.path.join(runs.get(self.run), f"{tag}.vis.db")
        self.repo = GraphRepoDB(db_path)
        self.conn = self.repo.get_db_connection()
        self.config_info = {}

    def load_graph_data(self):
        if not self.repo:
            return {'success': False, 'error': GraphUtils.t('dbInitError')}
        if not self.conn:
            self.conn = self.repo.get_db_connection()
        GraphState.set_global_value("all_node_info_cache", {})  # 切换文件清缓存
        return {'success': True}

    def load_graph_config_info(self):
        try:
            self.config_info = self.repo.query_config_info()
            GraphState.set_global_value("config_info", self.config_info)
            
            # 读取目录下配置文件列表
            modify_matched_files = GraphUtils.find_config_files(self.run)
            self.config_info['matchedConfigFiles'] = modify_matched_files or []
            return {'success': True, 'data': self.config_info}
        except Exception as e:
            logger.error(f"load graph config info failed, {e}")
            return {'success': False, 'error': GraphUtils.t('loadConfigError')}

    def load_graph_all_node_list(self, meta_data):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            micro_step = meta_data.get('microStep')
            if rank is None or step is None or micro_step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            result = {}
            if not self.config_info:
                self.config_info = self.repo.query_config_info() 
            # 单图
            if self.config_info.get('isSingleGraph'):
                # DB：根据graphType rank step micro_step 查询节点列表
                node_name_list = self.repo.query_node_name_list(rank, step, micro_step)
                result['npuNodeList'] = node_name_list
                return {'success': True, 'data': result} 
            # 双图
            else:
                config_data = GraphState.get_global_value("config_data")
                all_node_info = self.repo.query_all_node_info_in_one(rank, step, micro_step)
                config_data['npuMatchNodes'] = all_node_info.get('npu_match_node')
                config_data['benchMatchNodes'] = all_node_info.get('bench_match_node')
                config_data['npuUnMatchNodes'] = all_node_info.get('npu_unmatch_node')
                config_data['benchUnMatchNodes'] = all_node_info.get('bench_unmatch_node')
        
                result['npuMatchNodes'] = all_node_info.get('npu_match_node')
                result['benchMatchNodes'] = all_node_info.get('bench_match_node')
                result['npuUnMatchNodes'] = all_node_info.get('npu_unmatch_node')
                result['benchUnMatchNodes'] = all_node_info.get('bench_unmatch_node')
                result['npuNodeList'] = all_node_info.get('npu_node_list')
                result['benchNodeList'] = all_node_info.get('bench_node_list')
                
                return {'success': True, 'data': result}
        except Exception as e:
            logger.error(f"load graph all node list failed, {e}")
            return {'success': False, 'error': GraphUtils.t('loadNodeError')}

    def change_node_expand_state(self, node_info, meta_data):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            graph_type = node_info.get('nodeType')
            node_name = node_info.get('nodeName')
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            micro_step = meta_data.get('microStep')
            if rank is None or step is None or micro_step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            # 单图
            rank_step = {
                'rank': rank,
                'step': step
            }
            if self.config_info.get('isSingleGraph'):
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, SINGLE, self.repo, micro_step,
                                                                     rank_step)
            # NPU
            elif graph_type == NPU:
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, NPU, self.repo, micro_step, rank_step)
            # 标杆
            elif graph_type == BENCH:
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, BENCH, self.repo, micro_step, rank_step)
            else:
                return {'success': True, 'data': {}}
            return {'success': True, 'data': hierarchy}
        except Exception as e:
            logger.error('node expand or collapse failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('expandNodeError'), 'data': None}

    def search_node_by_precision(self, meta_data, values):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            micro_step = meta_data.get('microStep')
            if rank is None or step is None or micro_step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            is_filter_unmatch_nodes = True if UN_MATCHED_VALUE in values else False
            if is_filter_unmatch_nodes:
                values.remove(UN_MATCHED_VALUE)
          
            update_precision_cache = GraphState.get_global_value("update_precision_cache", {})
            node_name_list = []
            # 先查全局缓存
            if update_precision_cache:
                for node_name, node_info in update_precision_cache.items():
                    if not node_info.get("is_leaf_nodes"):
                        continue 
                    matched_node_link = node_info.get('matched_node_link', None)
                    if is_filter_unmatch_nodes and (not matched_node_link):
                        node_name_list.append(node_name)
                        
                    if isinstance(node_info.get("precision_index"), (int, float, complex)) and any(low <= node_info.
                                            get("precision_index", -1) <= high for low, high in values):
                        node_name_list.append(node_name)
            # 在查数据库
            else:
                node_name_list = self.repo.query_node_list_by_precision(step, rank, micro_step, values,
                                                                        is_filter_unmatch_nodes)
            return {'success': True, 'data': node_name_list}
        except Exception as e:
            logger.error('node search by precision failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('searchNodeError'), 'data': None}
    
    def search_node_by_overflow(self, meta_data, values):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            micro_step = meta_data.get('microStep')
            if rank is None or step is None or micro_step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            node_name_list = self.repo.query_node_list_by_overflow(step, rank, micro_step, values)
            return {'success': True, 'data': node_name_list}
        except Exception as e:
            logger.error('node search by overflow failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('searchNodeError'), 'data': None}

    def update_hierarchy_data(self, graph_type):
        if (graph_type == NPU or graph_type == BENCH):
            hierarchy = LayoutHierarchyModel.update_hierarchy_data(graph_type)
            return {'success': True, 'data': hierarchy}
        else:
            return {'success': False, 'error': GraphUtils.t('graphTypeError')}

    def get_node_info(self, node_info, meta_data):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            result = {}
            graph_type = node_info.get('nodeType')
            node_name = node_info.get('nodeName')
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            if rank is None or step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            if self.config_info.get('isSingleGraph') or graph_type == SINGLE:
                result['npu'] = self.repo.query_node_info(node_name, graph_type, rank, step)
            else:
                matched_node_type = BENCH if graph_type == NPU else NPU
                node = self.repo.query_node_info(node_name, graph_type, rank, step)
                matched_node_link = node.get('matched_node_link', []) 
                if isinstance(node.get('matched_node_link', []), list) and len(matched_node_link) > 0:
                    matched_node = self.repo.query_node_info(matched_node_link[-1], matched_node_type, rank, step)
                else:
                    matched_node = None
                result['npu'] = node if graph_type == NPU else matched_node
                result['bench'] = node if graph_type == BENCH else matched_node
            return {'success': True, 'data': result}
        except Exception as e:
            logger.error('get node info failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('getNodeInfoError'), 'data': None}

    def add_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_match_children):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            if rank is None or step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            task = self.config_info.get('task')
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                if is_match_children:
                    graph_data = self.repo.query_node_and_sub_nodes(npu_node_name, bench_node_name, rank, step)
                    match_result = MatchNodesController.process_task_add_child_layer(graph_data, npu_node_name,
                                                                               bench_node_name, task)
                else:
                    graph_data = self.repo.query_matched_nodes_info(npu_node_name, bench_node_name, rank, step)
                    match_result = MatchNodesController.process_task_add(graph_data, npu_node_name,
                                                                         bench_node_name, task)
                return self._generate_matched_result(match_result, rank, step)
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            logger.error(str(e))
            return {'success': False, 'error': GraphUtils.t('matchNodeError'), 'data': None}

    def add_match_nodes_by_config(self, config_file_name, meta_data):
        if not self.conn:
            return {'success': False, 'error': GraphUtils.t('dbInitError')}
        result = {}
        rank = meta_data.get('rank')
        step = meta_data.get('step')
        if rank is None or step is None:
            return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
        task = self.config_info.get('task')
        try:
            match_node_links, error = GraphUtils.safe_load_data(meta_data.get('run'), config_file_name)
            graph_data = self.repo.query_matched_nodes_info_by_config(match_node_links, rank, step)
            if error:
                return {'success': False, 'error': GraphUtils.t('loadConfigFileError')}
            if not match_node_links:
                return {'success': False, 'error': GraphUtils.t('matchNodeLinksNullError')}
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                match_result = MatchNodesController.process_task_add_child_layer_by_config(graph_data,
                                                                                           match_node_links, task)
                result = self._generate_matched_result(match_result, rank, step)
                result.setdefault('data', {})['matchResult'] = [item.get('success', False) for item in match_result]
                return result
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            logger.error(str(e))
            return {'success': False, 'error': GraphUtils.t('matchNodeError'), 'data': None}

    def delete_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_unmatch_children):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbInitError')}
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            if rank is None or step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            task = self.config_info.get('task')
            
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                if is_unmatch_children:
                    # DB：当前节点及其所有的子节点信息
                    graph_data = self.repo.query_node_and_sub_nodes(npu_node_name, bench_node_name, rank, step)
                    match_result = MatchNodesController.process_task_delete_child_layer(graph_data, npu_node_name,
                                                                                  bench_node_name, task)
                else:
                    graph_data = self.repo.query_matched_nodes_info(npu_node_name, bench_node_name, rank, step)
                    match_result = MatchNodesController.process_task_delete(graph_data, npu_node_name,
                                                                            bench_node_name, task)
                return self._generate_matched_result(match_result, rank, step)
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            logger.error('delete_match_nodes error: {}'.format(e))
            return {'success': False, 'error': GraphUtils.t('deleteMatchNodeFailed'), 'data': None}
    
    def update_precision_error(self, meta_data, filter_value):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbConncetionNotInit')}

            rank = meta_data.get('rank')
            step = meta_data.get('step')
            if rank is None or step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            npu_node_list = self.repo.query_node_info_by_data_source(step, rank, NPU)
            update_data_hierarchy = {}
            update_data_db = []
          
            for _, node_info in npu_node_list.items():
                output_statistical_diff = node_info.get('output_data', None)                
                if not node_info.get('matched_node_link') or not output_statistical_diff:
                    update_data_hierarchy[node_info.get('node_name')] = {
                       "precision_index": node_info.get('data').get('precision_index'),
                       "node_name": node_info.get('node_name'),
                       "matched_node_link": node_info.get('matched_node_link'),
                       'is_leaf_nodes': node_info.get('subnodes', []) == [],
                       'graph_type': NPU
                    }  
                    continue
                max_rel_error = -1
                #  根据filter_value 的选择指标计算新的误差值
                for _, diff_values in output_statistical_diff.items():
                    filter_diff_rel = []
                    if MAX_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MaxRelativeErr'))
                    if MIN_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MinRelativeErr'))
                    if NORM_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('NormRelativeErr'))
                    if MEAN_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MeanRelativeErr'))
                    # 过滤掉N/A
                    filter_diff_rel = [x for x in filter_diff_rel if x and x != 'N/A']
                    # 如果output指标中存在 Nan/inf/-inf, 直接标记为最大值
                    if "Nan" in filter_diff_rel or "inf" in filter_diff_rel or "-inf" in filter_diff_rel:
                        max_rel_error = 1
                        break
                    filter_diff_rel = [GraphUtils.convert_to_float(x) for x in filter_diff_rel]
                    max_rel_error_for_key = max(filter_diff_rel) if filter_diff_rel else 0
                    max_rel_error = max(max_rel_error, max_rel_error_for_key)
                if max_rel_error != -1:
                    update_data_db.append((
                        min(max_rel_error, 1),
                        step,
                        rank,
                        node_info.get('node_name')
                    ))
                update_data_hierarchy[node_info.get('node_name')] = {
                       "precision_index": min(max_rel_error, 1),
                       "node_name": node_info.get('node_name'),
                       "matched_node_link": node_info.get('matched_node_link'),
                       'is_leaf_nodes': node_info.get('subnodes', []) == [],
                       'graph_type': NPU
                    }  
            if len(update_data_hierarchy) > 0:
                # 视图：调用更新update_hirarchy方法，同步更新图
                LayoutHierarchyModel.update_current_hierarchy_data(list(update_data_hierarchy.values()))
                # 更新全局缓存，使用update_data_hierarchy覆盖原来的缓存即可，注意，切换视图需要重置缓存信息
                GraphState.set_global_value("update_precision_cache", update_data_hierarchy)
                return {'success': True, 'data': {}}
            else:
                return {'success': False, 'error': GraphUtils.t('noUpdateNodesFound')}
        except Exception as e:
            logger.error('update_precision_error error: {}'.format(e))
            return {'success': False, 'error': GraphUtils.t('updateNodeAccuracyFailed'), 'data': None}   

    def update_colors(self, colors):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbConncetionNotInit')}
            # DB：更新颜色
            update_db_res = self.repo.update_config_colors(colors)
            if not update_db_res:
                return {'success': False, 'error': GraphUtils.t('updateDatabaseFailed')}
            return {'success': True}
        except Exception as e:
            return {'success': False, 'error': GraphUtils.t('updateColorFailed'), 'data': None}

    def save_matched_relations(self, meta_data):
        try:
            if not self.conn:
                return {'success': False, 'error': GraphUtils.t('dbConncetionNotInit')}
            run = meta_data.get('run')
            tag = meta_data.get('tag')
            rank = meta_data.get('rank')
            step = meta_data.get('step')
            if rank is None or step is None:
                return {'success': False, 'error': GraphUtils.t('rankStepNullError')}
            # DB：根据step rank modify match_node_link查询已经修改的匹配成功的节点关系
            modify_matched_nodes_list = self.repo.query_modify_matched_nodes_list(rank, step)
            confilg_file_name = f"{tag}_{step}_{rank}.vis.config"
            _, error = GraphUtils.safe_save_data(modify_matched_nodes_list, run, confilg_file_name)
            if error:
                return {'success': False, 'error': error}
            else:
                return {'success': True, 'data': confilg_file_name}
       
        except Exception as e:
            logger.error('save_matched_relations error: {}'.format(e))
            return {'success': False, 'error': GraphUtils.t('saveMatchRelationshipFailed'), 'data': None}

    def _generate_matched_result(self, match_result, rank, step):
        update_data = []
        for item in match_result:
            if item.get('success') is True:
                nodes = item.get('data', [])
                for node in nodes:
                    update_data.append(node)
        if len(update_data) > 0:
            # DB：更新数据库节点信息
            update_db_res = self.repo.update_nodes_info(update_data, rank, step)
            if not update_db_res:
                return {'success': False, 'error': GraphUtils.t('updateDatabaseFailed')}
            # 视图：调用更新update_hirarchy方法，同步更新图
            LayoutHierarchyModel.update_current_hierarchy_data(update_data)
            # 返回：返回更新后的节点信息
            config_data = GraphState.get_global_value("config_data")
            result = {
                'success': True,
                'data': {
                    'npuMatchNodes': config_data.get('npuMatchNodes', {}),
                    'benchMatchNodes': config_data.get('benchMatchNodes', {}),
                    'npuUnMatchNodes': config_data.get('npuUnMatchNodes', []),
                    'benchUnMatchNodes': config_data.get('benchUnMatchNodes', [])
                }
            }
        else:
            result = {'success': False, 'error': GraphUtils.t('selectedNodeCantMatched')}
        return result
                     



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import abc

import numpy as np

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException, format_value


def handle_inf_nan(n_value, b_value):
    def convert_to_float(value):
        try:
            if isinstance(value, np.ndarray):
                return value.astype(float)
            else:
                return float(value)
        except ValueError as e:
            logger.error('\n'.join(e.args))
            raise CompareException(CompareException.INVALID_DATA_ERROR) from e

    n_value_convert, b_value_convert = convert_to_float(n_value), convert_to_float(b_value)
    """处理inf和nan的数据"""
    n_inf = np.isinf(n_value_convert)
    b_inf = np.isinf(b_value_convert)
    n_nan = np.isnan(n_value_convert)
    b_nan = np.isnan(b_value_convert)
    n_invalid = np.any(n_inf) or np.any(n_nan)
    b_invalid = np.any(b_inf) or np.any(b_nan)
    if n_invalid or b_invalid:
        if np.array_equal(n_inf, b_inf) and np.array_equal(n_nan, b_nan):
            n_value[n_inf] = 0
            b_value[b_inf] = 0
            n_value[n_nan] = 0
            b_value[b_nan] = 0
        else:
            return CompareConst.NAN, CompareConst.NAN
    return n_value, b_value


def get_error_flag_and_msg(n_value, b_value, error_flag=False, error_file=None):
    """判断数据是否有异常并返回异常的n_value, b_value，同时返回error_flag和error_msg"""
    err_msg = ""
    if error_flag:
        err_msg = f"Dump file: {error_file} not found or read failed."
        return CompareConst.READ_NONE, CompareConst.READ_NONE, error_flag, err_msg

    if n_value.size == 0:  # 判断读取到的数据是否为空
        err_msg = "This is empty data, can not compare."
        error_flag = True
        return CompareConst.NONE, CompareConst.NONE, error_flag, err_msg
    if not n_value.shape:  # 判断数据是否为0维张量
        err_msg = (f"This is type of 0-d tensor, can not calculate '{CompareConst.COSINE}', '{CompareConst.EUC_DIST}', "
                   f"'{CompareConst.ONE_THOUSANDTH_ERR_RATIO}' and '{CompareConst.FIVE_THOUSANDTHS_ERR_RATIO}'. ")
        error_flag = False  # 0-d tensor 最大绝对误差、最大相对误差仍然支持计算，因此error_flag设置为False，不做统一处理
        return n_value, b_value, error_flag, err_msg
    if n_value.shape != b_value.shape:  # 判断NPU和bench的数据结构是否一致
        err_msg = "Shape of NPU and bench tensor do not match. Skipped."
        error_flag = True
        return CompareConst.SHAPE_UNMATCH, CompareConst.SHAPE_UNMATCH, error_flag, err_msg

    try:
        n_value, b_value = handle_inf_nan(n_value, b_value)  # 判断是否有nan/inf数据
    except CompareException:
        logger.error('Numpy data is unreadable, please check!')
        err_msg = "Data is unreadable."
        error_flag = True
        return CompareConst.UNREADABLE, CompareConst.UNREADABLE, error_flag, err_msg
    if n_value is CompareConst.NAN or b_value is CompareConst.NAN:
        err_msg = "The position of inf or nan in NPU and bench Tensor do not match."
        error_flag = True
        return CompareConst.NAN, CompareConst.NAN, error_flag, err_msg

    if n_value.dtype != b_value.dtype:  # 判断数据的dtype是否一致
        err_msg = "Dtype of NPU and bench tensor do not match."
        error_flag = False
        return n_value, b_value, error_flag, err_msg

    return n_value, b_value, error_flag, err_msg


def reshape_value(n_value, b_value):
    """返回reshape后的数据"""
    if not n_value.shape:  # 判断数据是否为0维tensor， 如果0维tensor，不会转成1维tensor，直接返回
        if n_value.dtype == bool:
            n_value = n_value.astype(float)
            b_value = b_value.astype(float)
        return n_value, b_value

    n_value = n_value.reshape(-1).astype(float)  # 32转64为了防止某些数转dataframe时出现误差
    b_value = b_value.reshape(-1).astype(float)
    return n_value, b_value


def npy_data_check(n_value, b_value):
    error_message = ""
    if not isinstance(n_value, np.ndarray) or not isinstance(b_value, np.ndarray):
        error_message += "Dump file is not ndarray.\n"

    # 检查 n_value 和 b_value 是否为空
    if not error_message and (n_value.size == 0 or b_value.size == 0):
        error_message += "This is empty data, can not compare.\n"

    if not error_message:
        if not n_value.shape or not b_value.shape:
            error_message += "This is type of scalar data, can not compare.\n"
        if n_value.shape != b_value.shape:
            error_message += "Shape of NPU and bench Tensor do not match.\n"
        if n_value.dtype != b_value.dtype:
            error_message += "Dtype of NPU and bench Tensor do not match. Skipped.\n"

    if not error_message:
        try:
            n_value, b_value = handle_inf_nan(n_value, b_value)  # 判断是否有nan/inf数据
        except CompareException:
            logger.error('Numpy data is unreadable, please check!')
            return True, 'Numpy data is unreadable, please check!'
        # handle_inf_nan 会返回'Nan'或ndarray类型，使用类型判断是否存在无法处理的nan/inf数据
        if not isinstance(n_value, np.ndarray) or not isinstance(b_value, np.ndarray):
            error_message += "The position of inf or nan in NPU and bench Tensor do not match.\n"
    if error_message == "":
        error_flag = False
    else:
        error_flag = True
    return error_flag, error_message


def statistics_data_check(result_dict):
    error_message = ""

    if result_dict.get(CompareConst.NPU_NAME) is None or result_dict.get(CompareConst.BENCH_NAME) is None:
        error_message += "Dump file not found.\n"

    if not result_dict.get(CompareConst.NPU_SHAPE) or not result_dict.get(CompareConst.BENCH_SHAPE):
        error_message += "This is type of scalar data, can not compare.\n"
    elif result_dict.get(CompareConst.NPU_SHAPE) != result_dict.get(CompareConst.BENCH_SHAPE):
        error_message += "Tensor shapes do not match.\n"

    if result_dict.get(CompareConst.NPU_DTYPE) != result_dict.get(CompareConst.BENCH_DTYPE):
        error_message += "Dtype of NPU and bench Tensor do not match. Skipped.\n"

    if error_message == "":
        error_flag = False
    else:
        error_flag = True
    return error_flag, error_message


class TensorComparisonBasic(abc.ABC):
    """NPU和bench中npy数据的比较模板"""

    @abc.abstractmethod
    def apply(self, n_value, b_value, relative_err, err_msg):
        raise NotImplementedError


def get_relative_err(n_value, b_value):
    """计算相对误差"""
    with np.errstate(divide='ignore', invalid='ignore'):
        if n_value.dtype not in CompareConst.FLOAT_TYPE:
            n_value = n_value.astype(float)
        if b_value.dtype not in CompareConst.FLOAT_TYPE:
            b_value = b_value.astype(float)

        n_value_copy = n_value.copy()
        b_value_copy = b_value.copy()
        zero_mask = (b_value_copy == 0)
        b_value_copy[zero_mask] += Const.FLOAT_EPSILON
        n_value_copy[zero_mask] += Const.FLOAT_EPSILON
        relative_err = np.divide((n_value_copy - b_value_copy), b_value_copy)
    return np.abs(relative_err)


class GetCosineSimilarity(TensorComparisonBasic):
    """计算cosine相似度"""

    @staticmethod
    def correct_data(result):
        if result == CompareConst.NAN:
            return result
        if float(result) > CompareConst.COSINE_THRESHOLD:
            return round(float(result), 6)
        return result

    def apply(self, n_value, b_value, relative_err, err_msg):
        if "This is type of 0-d tensor" in err_msg:
            return CompareConst.UNSUPPORTED, err_msg

        with np.errstate(divide="ignore", invalid="ignore"):
            if len(n_value) == 1:
                return CompareConst.UNSUPPORTED, "This is a 1-d tensor of length 1."
            num = n_value.dot(b_value)
            a_norm = np.linalg.norm(n_value)
            b_norm = np.linalg.norm(b_value)

            if a_norm <= Const.FLOAT_EPSILON and b_norm <= Const.FLOAT_EPSILON:
                return 1.0, ""
            if a_norm <= Const.FLOAT_EPSILON:
                return CompareConst.NAN, "Cannot compare by Cosine Similarity, All the data is Zero in npu dump data."
            if b_norm <= Const.FLOAT_EPSILON:
                return CompareConst.NAN, "Cannot compare by Cosine Similarity, All the data is Zero in Bench dump data."

            cos = num / (a_norm * b_norm)
            if np.isnan(cos):
                return CompareConst.NAN, "Cannot compare by Cosine Similarity, the dump data has NaN."
            result = format_value(cos)
            result = self.correct_data(result)
        return result, ""


class GetEuclideanDistance(TensorComparisonBasic):
    """计算欧式距离"""

    def apply(self, n_value, b_value, relative_err, err_msg):
        if "This is type of 0-d tensor" in err_msg:
            return CompareConst.UNSUPPORTED, err_msg

        distance = np.linalg.norm(n_value - b_value, ord=2)

        return distance, ""


class GetMaxAbsErr(TensorComparisonBasic):
    """计算最大绝对误差"""

    def apply(self, n_value, b_value, relative_err, err_msg):
        temp_res = n_value - b_value
        max_value = np.max(np.abs(temp_res))
        if np.isnan(max_value):
            msg = "Cannot compare by MaxAbsError, the data contains nan/inf/-inf in dump data."
            return CompareConst.NAN, msg
        return format_value(max_value), ""


class GetMaxRelativeErr(TensorComparisonBasic):
    """计算最大相对误差"""

    def apply(self, n_value, b_value, relative_err, err_msg):
        max_relative_err = np.max(np.abs(relative_err))
        if np.isnan(max_relative_err):
            msg = "Cannot compare by MaxRelativeError, the data contains nan/inf/-inf in dump data."
            return CompareConst.NAN, msg
        return format_value(max_relative_err), ""


class GetErrRatio(TensorComparisonBasic):
    """计算相对误差小于指定阈值(千分之一、千分之五)的比例"""

    def __init__(self, threshold):
        self.threshold = threshold

    def apply(self, n_value, b_value, relative_err, err_msg):
        if "This is type of 0-d tensor" in err_msg:
            return CompareConst.UNSUPPORTED, err_msg

        if not np.size(relative_err):
            return CompareConst.NAN, ""

        ratio = np.sum(relative_err < self.threshold) / np.size(relative_err)
        return format_value(ratio), ""


class CompareOps:
    compare_ops = {
        "cosine_similarity": GetCosineSimilarity(),
        "euclidean_distance": GetEuclideanDistance(),
        "max_abs_error": GetMaxAbsErr(),
        "max_relative_error": GetMaxRelativeErr(),
        "one_thousand_err_ratio": GetErrRatio(CompareConst.THOUSAND_RATIO_THRESHOLD),
        "five_thousand_err_ratio": GetErrRatio(CompareConst.FIVE_THOUSAND_RATIO_THRESHOLD)
    }


def error_value_process(n_value):
    if n_value in [CompareConst.READ_NONE, CompareConst.UNREADABLE, CompareConst.NONE,
                   CompareConst.NO_REAL_DATA, CompareConst.API_UNMATCH]:
        return CompareConst.UNSUPPORTED, ""
    if n_value == CompareConst.SHAPE_UNMATCH:
        return CompareConst.SHAPE_UNMATCH, ""
    if n_value == CompareConst.NAN:
        return CompareConst.N_A, ""
    return CompareConst.N_A, ""


def compare_ops_apply(n_value, b_value, error_flag, err_msg):
    result_list = []
    if error_flag:
        result, msg = error_value_process(n_value)
        result_list = [result] * len(CompareOps.compare_ops)
        err_msg += msg
        return result_list, err_msg

    relative_err = get_relative_err(n_value, b_value)
    n_value, b_value = reshape_value(n_value, b_value)

    for op in CompareOps.compare_ops.values():
        result, msg = op.apply(n_value, b_value, relative_err, err_msg)
        result_list.append(result)
        err_msg += msg
    return result_list, err_msg



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Make advisor, perform comparative analysis, This class mainly involves the main function.
"""

import pandas as pd

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.advisor.advisor_const import AdvisorConst
from msprobe.msaccucmp.advisor.advisor_result import AdvisorResult
from msprobe.msaccucmp.advisor.input_advisor import InputAdvisor
from msprobe.msaccucmp.advisor.node_advisor import NodeAdvisor
from msprobe.msaccucmp.advisor.overflow_advisor import OverflowAdvisor


class CompareAdvisor:
    """
    Class for generate advisor
    """

    def __init__(self, input_file, input_nodes=None, out_path=""):
        self.input_file = input_file
        self.input_nodes = input_nodes
        self.out_path = out_path

    @staticmethod
    def _overflow_check(advisor_result, analyze_data):
        if not advisor_result.match_advisor:
            overflow_advisor = OverflowAdvisor(analyze_data, advisor_result)
            advisor_result = overflow_advisor.start_analyze()
            if advisor_result.match_advisor:
                log.print_info_log("The FP16 Overflow detection matches successfully.")
            log.print_info_log("End FP16 Overflow detection.")
        return advisor_result

    @staticmethod
    def _net_nodes_check(advisor_result, analyze_data):
        if not advisor_result.match_advisor:
            node_advisor = NodeAdvisor(analyze_data, advisor_result)
            advisor_result = node_advisor.start_analyze()
            if advisor_result.match_advisor:
                log.print_info_log("The Global Consistency detection matches successfully.")
            log.print_info_log("End Global Consistency detection.")
        return advisor_result

    def advisor(self):
        analyze_data = self._parse_input_file()
        log.print_info_log('Start analyzing the comparison results: "%r" .' % self.input_file)
        advisor_result = AdvisorResult()
        advisor_result = self._overflow_check(advisor_result, analyze_data)
        advisor_result = self._input_check(advisor_result, analyze_data)
        advisor_result = self._net_nodes_check(advisor_result, analyze_data)
        log.print_info_log("Comparison result analysis is over.")
        return advisor_result

    def _input_check(self, advisor_result, analyze_data):
        if not advisor_result.match_advisor and self.input_nodes:
            input_advisor = InputAdvisor(analyze_data, advisor_result, self.input_nodes)
            advisor_result = input_advisor.start_analyze()
            if advisor_result.match_advisor:
                log.print_info_log("The Input Inconsistent detection matches successfully.")
            log.print_info_log("End Input Inconsistent detection.")
        return advisor_result

    def _parse_input_file(self):
        if self.input_file.endswith(".csv"):
            try:
                df = pd.read_csv(self.input_file, on_bad_lines='skip')
            except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as io_err:
                log.print_error_log('Failed to parse the input file %r. %s'
                                    % (self.input_file, str(io_err)))
                raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from io_err
            data_columns = df.columns.values
            if not {AdvisorConst.INDEX, AdvisorConst.NPU_DUMP}.issubset(data_columns):
                log.print_error_log('Input csv file does not contain %s, %s columns.'
                                    % (AdvisorConst.INDEX, AdvisorConst.NPU_DUMP))
                raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)
            return df
        else:
            log.print_error_log("Advisor only support csv file from msaccucmp result.")
            raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import CompareConst
from msprobe.core.common.file_utils import logger
from msprobe.core.common.utils import CompareException


def replace_compare_index_dict(compare_index_dict, compare_index_list, rank_num):
    """
    比对指标值为N/A、unsupported、Nan，将比对指标值替换成NPU max 和 Bench max(几个统计量相同)

    示例：
    Distributed.all_reduce.0.forward.output.group的比对指标值是N/A
    替换后：
    比对指标值为:
        NPU: tp-0-1-2-3
        Bench: tp-0-1-2-3
    """

    if CompareConst.NPU_MAX not in compare_index_dict or CompareConst.BENCH_MAX not in compare_index_dict:
        compare_index_dict.pop(CompareConst.NPU_MAX, None)
        compare_index_dict.pop(CompareConst.BENCH_MAX, None)
        return compare_index_dict

    # 遍历比对指标列表，排除最后两个指标NPU max， Bench max
    for compare_index in compare_index_list[:-2]:
        op_name_index_dict = compare_index_dict[compare_index]
        # 遍历op_item名称和对应的比对指标值
        for op_name, index_value in op_name_index_dict.items():
            npu_max = compare_index_dict[CompareConst.NPU_MAX][op_name][rank_num]
            bench_max = compare_index_dict[CompareConst.BENCH_MAX][op_name][rank_num]
            # 如果当前比对指标值是N/A、unsupported、Nan，并且NPU和Bench的最大值是类型相同，进行替换
            if index_value[rank_num] in [CompareConst.N_A, CompareConst.UNSUPPORTED, CompareConst.NAN]:
                compare_index_dict[compare_index][op_name][rank_num] = f'NPU:{str(npu_max)}  Bench:{str(bench_max)}'

    # 删除NPU_MAX和BENCH_MAX
    compare_index_dict.pop(CompareConst.NPU_MAX, None)
    compare_index_dict.pop(CompareConst.BENCH_MAX, None)
    return compare_index_dict


def check_config(config):
    """
    config.yaml 内容检查
    Args: config:
    Returns: config
    """
    if not config:
        logger.error('config.yaml is empty, please check.')
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)

    api_list = config.get('api')
    if not api_list:
        logger.error('The APIs required to merge data were not found.')
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)
    if not isinstance(api_list, list):
        logger.error("The config format of 'api' is incorrect, please check.")
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)

    compare_index_list = config.get('compare_index', [])
    if compare_index_list is None:
        compare_index_list = []
        config['compare_index'] = compare_index_list
    if not isinstance(compare_index_list, list):
        logger.error("The config format of 'compare_index' is incorrect, please check.")
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)

    return config
